import os
import posixpath
import shutil
import warnings
from collections import ChainMap
from collections import OrderedDict
from collections.abc import ItemsView
from collections.abc import KeysView
from collections.abc import Mapping
from collections.abc import MutableMapping
from collections.abc import ValuesView
from contextlib import suppress
from functools import wraps
from itertools import chain

from lektor.constants import PRIMARY_ALT
from lektor.metaformat import serialize
from lektor.utils import atomic_open
from lektor.utils import cleanup_path
from lektor.utils import increment_filename
from lektor.utils import is_valid_id
from lektor.utils import parse_path
from lektor.utils import secure_filename


implied_keys = {"_id", "_path", "_gid", "_alt", "_source_alt", "_attachment_for"}
possibly_implied_keys = {"_model", "_template", "_attachment_type"}


class BadEdit(Exception):
    pass


class BadDelete(BadEdit):
    pass


def _is_valid_path(path: str) -> bool:
    split_path = path.strip("/").split("/")
    if split_path == [""]:
        split_path = []
    return parse_path(path) == split_path


def make_editor_session(pad, path, is_attachment=None, alt=PRIMARY_ALT, datamodel=None):
    """Creates an editor session for the given path object."""
    if not _is_valid_path(path):
        raise BadEdit("Invalid path")
    path = cleanup_path(path)

    if alt != PRIMARY_ALT and not pad.db.config.is_valid_alternative(alt):
        raise BadEdit(f"Attempted to edit an invalid alternative ({alt})")

    raw_data = pad.db.load_raw_data(path, cls=OrderedDict, alt=alt, fallback=False)
    raw_data_fallback = None
    if alt != PRIMARY_ALT:
        raw_data_fallback = pad.db.load_raw_data(path, cls=OrderedDict)
        all_data = OrderedDict()
        all_data.update(raw_data_fallback or ())
        all_data.update(raw_data or ())
    else:
        all_data = raw_data

    id = posixpath.basename(path)
    if not is_valid_id(id):
        raise BadEdit("Invalid ID")

    record = None
    exists = raw_data is not None or raw_data_fallback is not None
    if raw_data is None:
        raw_data = OrderedDict()

    if is_attachment is None:
        if not exists:
            is_attachment = False
        else:
            is_attachment = bool(all_data.get("_attachment_for"))
    elif bool(all_data.get("_attachment_for")) != is_attachment:
        raise BadEdit(
            "The attachment flag passed is conflicting with the "
            "record's attachment flag."
        )

    if exists:
        # XXX: what about changing the datamodel after the fact?
        if datamodel is not None:
            raise BadEdit(
                "When editing an existing record, a datamodel must not be provided."
            )
        datamodel = pad.db.get_datamodel_for_raw_data(all_data, pad)
    else:
        if datamodel is None:
            datamodel = pad.db.get_implied_datamodel(path, is_attachment, pad)
        elif isinstance(datamodel, str):
            datamodel = pad.db.datamodels[datamodel]

    if exists:
        record = pad.instance_from_data(dict(all_data), datamodel)

    for key in implied_keys:
        raw_data.pop(key, None)
        if raw_data_fallback:
            raw_data_fallback.pop(key, None)

    return EditorSession(
        pad,
        id,
        str(path),
        original_data=raw_data,
        fallback_data=raw_data_fallback,
        datamodel=datamodel,
        record=record,
        exists=exists,
        is_attachment=is_attachment,
        alt=alt,
    )


def _deprecated_data_proxy(wrapped):
    """Issue warning when deprecated mapping methods are used directly on
    EditorSession.
    """

    name = wrapped.__name__
    newname = name.removeprefix("iter")

    @wraps(wrapped)
    def wrapper(self, *args, **kwargs):
        warnings.warn(
            f"EditorSession.{name} has been deprecated as of Lektor 3.3.2. "
            f"Please use EditorSession.data.{newname} instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return wrapped(self, *args, **kwargs)

    return wrapper


class EditorSession:
    def __init__(
        self,
        pad,
        id,
        path,
        *,
        original_data,
        fallback_data,
        datamodel,
        record,
        exists=True,
        is_attachment=False,
        alt=PRIMARY_ALT,
    ):
        self.id = id
        self.pad = pad
        self.path = path
        self.record = record
        self.exists = exists
        self.datamodel = datamodel
        self.is_root = path.strip("/") == ""
        self.alt = alt

        slug_format = None
        parent_name = posixpath.dirname(path)
        if parent_name != path:
            parent = pad.get(parent_name)
            if parent is not None:
                slug_format = parent.datamodel.child_config.slug_format
        if slug_format is None:
            slug_format = "{{ this._id }}"
        self.slug_format = slug_format
        self.implied_attachment_type = None

        if is_attachment:
            self.implied_attachment_type = pad.db.get_attachment_type(path)

        self.data = MutableEditorData(original_data, fallback_data)

        self._delete_this = False
        self._recursive_delete = False
        self._master_delete = False
        self.is_attachment = is_attachment
        self.closed = False

    def to_json(self):
        label = None
        label_i18n = None
        url_path = None
        if self.record is not None:
            label = self.record.record_label
            label_i18n = self.record.get_record_label_i18n()
            url_path = self.record.url_path
        else:
            label = self.id
        can_be_deleted = not self.datamodel.protected and not self.is_root
        return {
            "data": dict(self.data.items()),
            "record_info": {
                "id": self.id,
                "path": self.path,
                "exists": self.exists,
                "label": label,
                "label_i18n": label_i18n,
                "url_path": url_path,
                "alt": self.alt,
                "is_attachment": self.is_attachment,
                "can_be_deleted": can_be_deleted,
                "slug_format": self.slug_format,
                "implied_attachment_type": self.implied_attachment_type,
                "default_template": self.datamodel.get_default_template_name(),
            },
            "datamodel": self.datamodel.to_json(self.pad, self.record, self.alt),
        }

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        if exc_type is not None:
            self.rollback()
        else:
            self.commit()

    def get_fs_path(self, alt=PRIMARY_ALT):
        """The path to the record file on the file system."""
        base = self.pad.db.to_fs_path(self.path)
        suffix = ".lr"
        if alt != PRIMARY_ALT:
            suffix = f"+{alt}{suffix}"
        if self.is_attachment:
            return base + suffix
        return os.path.join(base, "contents" + suffix)

    @property
    def fs_path(self):
        """The file system path of the content file on disk."""
        return self.get_fs_path(self.alt)

    @property
    def attachment_fs_path(self):
        """The file system path of the actual attachment."""
        if self.is_attachment:
            return self.pad.db.to_fs_path(self.path)
        return None

    def rollback(self):
        """Ignores all changes and rejects them."""
        if self.closed:
            return
        self.closed = True

    def commit(self):
        """Saves changes back to the file system."""
        if not self.closed:
            if self._delete_this:
                self._delete_impl()
            else:
                self._save_impl()
        self.closed = True

    def delete(self, recursive=None, delete_master=False):
        """Deletes the record.  How the delete works depends on what is being
        deleted:

        *   delete attachment: recursive mode is silently ignored.  If
            `delete_master` is set then the attachment is deleted, otherwise
            only the metadata is deleted.
        *   delete page: in recursive mode everything is deleted in which
            case `delete_master` must be set to `True` or an error is
            generated.  In fact, the default is to perform a recursive
            delete in that case.  If `delete_master` is False, then only the
            contents file of the current alt is deleted.

        If a delete cannot be performed, an error is generated.
        """
        if self.closed:
            return
        if recursive is None:
            recursive = not self.is_attachment and delete_master
        self._delete_this = True
        self._recursive_delete = recursive
        self._master_delete = delete_master

    def add_attachment(self, filename, fp):
        """Stores a new attachment.  Returns `None` if the file already"""
        if not self.exists:
            raise BadEdit("Record does not exist.")
        if self.is_attachment:
            raise BadEdit("Cannot attach something to an attachment.")
        if not self.datamodel.has_own_attachments:
            raise BadEdit("Attachments are disabled for this page.")
        directory = self.pad.db.to_fs_path(self.path)

        safe_filename = secure_filename(filename)

        while 1:
            fn = os.path.join(directory, safe_filename)
            if not os.path.isfile(fn):
                break
            safe_filename = increment_filename(fn)

        with atomic_open(fn, "wb") as f:
            shutil.copyfileobj(fp, f)
        return safe_filename

    def _attachment_delete_impl(self):
        files = [self.fs_path]
        if self._master_delete:
            files.append(self.attachment_fs_path)
            files.extend(
                self.get_fs_path(alt) for alt in self.pad.db.config.list_alternatives()
            )
        for fn in files:
            try:
                os.unlink(fn)
            except OSError:  # noqa: PERF203
                pass

    def _page_delete_impl(self):
        directory = os.path.dirname(self.fs_path)

        if self._recursive_delete:
            with suppress(OSError):
                shutil.rmtree(directory)
            return
        if self._master_delete:
            raise BadDelete(
                "Master deletes of pages require that recursive deleting is enabled."
            )

        for fn in self.fs_path, directory:
            try:
                os.unlink(fn)
            except OSError:  # noqa: PERF203
                pass

    def _delete_impl(self):
        if self.alt != PRIMARY_ALT:
            if self._master_delete:
                raise BadDelete(
                    "Master deletes need to be done from the primary "
                    f'alt.  Tried to delete from "{self.alt}"'
                )
            if self._recursive_delete:
                raise BadDelete(
                    "Cannot perform recursive delete from a non "
                    f'primary alt.  Tried to delete from "{self.alt}"'
                )

        if self.is_attachment:
            self._attachment_delete_impl()
        else:
            self._page_delete_impl()

    def _save_impl(self):
        # When creating a new alt from a primary self.exists is True but
        # the file does not exist yet.  In this case we want to explicitly
        # create it anyways instead of bailing.
        if not self.data.ischanged() and self.exists and os.path.exists(self.fs_path):
            return

        try:
            os.makedirs(os.path.dirname(self.fs_path))
        except OSError:
            pass

        with atomic_open(self.fs_path, "wb") as f:
            for chunk in serialize(self.data.items(fallback=False), encoding="utf-8"):
                f.write(chunk)

    def __repr__(self):
        bits = [repr(self.path)]
        if self.alt != PRIMARY_ALT:
            bits.append(f" alt={self.alt!r}")
        if not self.exists:
            bits.append(" new")

        return f"<{self.__class__.__name__} {' '.join(bits)}>"

    # The mapping methods used to access the page data have been moved
    # to EditorSession.data.
    #
    # We have left behind these proxy methods so as not to break any existing
    # external code.
    @_deprecated_data_proxy
    def revert_key(self, key):
        self.data.revert_key(key)

    @_deprecated_data_proxy
    def __contains__(self, key):
        return key in self.data

    @_deprecated_data_proxy
    def __getitem__(self, key):
        return self.data[key]

    @_deprecated_data_proxy
    def __len__(self):
        return len(self.data)

    @_deprecated_data_proxy
    def __iter__(self):
        return iter(self.data)

    @_deprecated_data_proxy
    def items(self, fallback=True):
        return self.data.items(fallback)

    @_deprecated_data_proxy
    def keys(self, fallback=True):
        return self.data.keys(fallback)

    @_deprecated_data_proxy
    def values(self, fallback=True):
        return self.data.values(fallback)

    @_deprecated_data_proxy
    def iteritems(self, fallback=True):
        return self.data.items(fallback)

    @_deprecated_data_proxy
    def iterkeys(self, fallback=True):
        return self.data.keys(fallback)

    @_deprecated_data_proxy
    def itervalues(self, fallback=True):
        return self.data.values(fallback)

    @_deprecated_data_proxy
    def __setitem__(self, key, value):
        self.data[key] = value

    @_deprecated_data_proxy
    def update(self, *args, **kwargs):
        self.data.update(*args, **kwargs)


del _deprecated_data_proxy


class EditorData(Mapping):
    """A read-only view of edited data.

    This is a chained dict with (possibly) mutated data overlaid on
    the original data for the record.
    """

    def __init__(self, original_data, fallback_data=None, _changed_data=None):
        if _changed_data is None:
            _changed_data = {}
        self.fallback_data = fallback_data
        if fallback_data:
            self._orig_data = ChainMap(original_data, fallback_data)
            self._data = ChainMap(_changed_data, original_data, fallback_data)
        else:
            self._orig_data = original_data
            self._data = ChainMap(_changed_data, original_data)

    @property
    def _changed_data(self):
        return self._data.maps[0]

    @property
    def original_data(self):
        return self._data.maps[1]

    def ischanged(self):
        return len(self._changed_data) > 0

    def __getitem__(self, key):
        rv = self._data.get(key)
        if rv is None:
            raise KeyError(key)
        return rv

    def __iter__(self):
        data = self._data
        fallback_data = self.fallback_data or {}

        for key in _uniq(chain(self.original_data, fallback_data, sorted(data))):
            if key not in implied_keys:
                if data[key] is not None:
                    yield key

    def __len__(self):
        data = self._data
        return sum(
            1 for key in data if key not in implied_keys and data[key] is not None
        )

    def keys(self, fallback=True):  # pylint: disable=arguments-differ
        return KeysView(self if fallback else self._without_fallback())

    def items(self, fallback=True):  # pylint: disable=arguments-differ
        return ItemsView(self if fallback else self._without_fallback())

    def values(self, fallback=True):  # pylint: disable=arguments-differ
        return ValuesView(self if fallback else self._without_fallback())

    def _without_fallback(self):
        """Return a copy of self, with fallback_data set to None."""
        if not self.fallback_data:
            return self
        return EditorData(self.original_data, _changed_data=self._changed_data)


class MutableEditorData(EditorData, MutableMapping):
    """A mutable view of edited data.

    This is a chained dict with (possibly) mutated data overlaid on
    the original data for the record.
    """

    def __setitem__(self, key, value):
        if key in implied_keys:
            raise KeyError(f"Can not set implied key {key!r}")
        orig_value = self._orig_data.get(key)
        if value != orig_value:
            self._data[key] = value
        elif key in possibly_implied_keys:
            # If the key is in the possibly implied key set and set to
            # that value, we will set it to changed anyways.  This allows
            # overriding of such special keys.
            self._data[key] = value
        else:
            self._data.pop(key, None)

    def __delitem__(self, key):
        if key in implied_keys or self._data.get(key) is None:
            raise KeyError(key)
        self[key] = None

    def revert_key(self, key):
        """Reverts a key to the implied value."""
        self._data.pop(key, None)


def _uniq(seq):
    """Return items from iterable in order, skipping items that have already been seen.

    The items in ``seq`` must be hashable.
    """
    seen = set()
    for item in seq:
        if item not in seen:
            seen.add(item)
            yield item



import os
import re
import sys

import pytest

from lektor.buildfailures import BuildFailure
from lektor.buildfailures import FailureController


@pytest.fixture
def output_path(tmp_path):
    return tmp_path.__fspath__()


@pytest.fixture
def failure_controller(pad, output_path):
    return FailureController(pad, output_path)


def test_BuildFailure_from_exc_info():
    def throw_exception():
        x = {}
        try:
            x["somekey"]
        except KeyError:
            raise RuntimeError("test error")  # pylint: disable=raise-missing-from # noqa: B904

    artifact_name = "test_artifact"
    failure = None
    try:
        throw_exception()
    except Exception:
        failure = BuildFailure.from_exc_info(artifact_name, sys.exc_info())

    assert failure
    assert failure.data["artifact"] == artifact_name
    assert failure.data["exception"] == "RuntimeError: test error"
    traceback = failure.data["traceback"]
    print(traceback)
    patterns = [
        r'x\["somekey"\]',
        r"KeyError: .somekey.",
        r"During handling of the above exception, another exception occurred",
        r"throw_exception\(\)",
        r'raise RuntimeError\("test error"\)',
        r"RuntimeError: test error",
    ]
    for pattern in patterns:
        assert re.search(pattern, traceback)


def test_failure_controller(failure_controller):
    try:
        raise RuntimeError("test exception")
    except Exception:
        failure_controller.store_failure("artifact_name", sys.exc_info())

    failure = failure_controller.lookup_failure("artifact_name")
    assert failure.data["exception"] == "RuntimeError: test exception"

    failure_controller.clear_failure("artifact_name")
    assert failure_controller.lookup_failure("artifact_name") is None


def test_failure_controller_clear_lookup_missing(failure_controller):
    assert failure_controller.lookup_failure("missing_artifact") is None


def test_failure_controller_clear_missing(failure_controller):
    failure_controller.clear_failure("missing_artifact")
    assert failure_controller.lookup_failure("missing_artifact") is None


def test_failure_controller_fs_exceptions(failure_controller):
    # Create a directory in the location that FailureController wants
    # a file stored to trigger unexpected OSErrors.
    filename = failure_controller.get_filename("broken")
    os.makedirs(filename)

    with pytest.raises(OSError):
        failure_controller.lookup_failure("broken")
    with pytest.raises(OSError):
        failure_controller.clear_failure("broken")
    with pytest.raises(OSError):
        try:
            raise RuntimeError("test exception")
        except Exception:
            failure_controller.store_failure("broken", sys.exc_info())



from pathlib import Path

import pytest

from lektor.cli_utils import ResolvedPath


@pytest.mark.parametrize("file_exists", [True, False])
def test_ResolvedPath_resolves_relative_path(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch, file_exists: bool
) -> None:
    resolved_path = ResolvedPath()
    monkeypatch.chdir(tmp_path)
    if file_exists:
        Path(tmp_path, "filename").touch()

    resolved = resolved_path.convert("filename", None, None)
    assert isinstance(resolved, str)
    assert resolved == str(Path.cwd().resolve() / "filename")



# pylint: disable=import-outside-toplevel
import os
import sys
import warnings
from importlib import metadata
from itertools import chain

import click

from lektor.cli_utils import AliasedGroup
from lektor.cli_utils import echo_json
from lektor.cli_utils import extraflag
from lektor.cli_utils import pass_context
from lektor.cli_utils import pruneflag
from lektor.cli_utils import ResolvedPath
from lektor.cli_utils import validate_language
from lektor.devcli import cli as devcli
from lektor.project import Project
from lektor.utils import secure_url


version = metadata.version("Lektor")


@click.group(cls=AliasedGroup)
@click.option(
    "--project",
    type=click.Path(exists=True),
    help="The path to the lektor project to work with.",
)
@click.option(
    "--language",
    default=None,
    callback=validate_language,
    help="The UI language to use (overrides autodetection).",
)
@click.version_option(prog_name="Lektor", version=version)
@pass_context
def cli(ctx, project=None, language=None):
    """The lektor management application.

    This command can invoke lektor locally and serve up the website.  It's
    intended for local development of websites.
    """
    if not sys.warnoptions:
        warnings.simplefilter("default")
    if language is not None:
        ctx.ui_lang = language
    if project is not None:
        ctx.set_project_path(project)


@cli.command("build")
@click.option(
    "-O",
    "--output-path",
    type=ResolvedPath(writable=True, file_okay=False),
    default=None,
    help="The output path.",
)
@click.option(
    "--watch",
    is_flag=True,
    help="If this is enabled the build "
    "process goes into an automatic loop where it watches the "
    "file system for changes and rebuilds.",
)
@pruneflag
@click.option(
    "-v",
    "--verbose",
    "verbosity",
    count=True,
    help="Increases the verbosity of the logging.",
)
@click.option(
    "--source-info-only",
    is_flag=True,
    help="Instead of building only updates the source infos.  The "
    "source info is used by the web admin panel to quickly find "
    "information about the source files (for instance jump to "
    "files).",
)
@click.option(
    "--buildstate-path",
    type=click.Path(writable=True, file_okay=False),
    default=None,
    help="Path to a directory that Lektor will use for coordinating "
    "the state of the build. Defaults to a directory named "
    "`.lektor` inside the output path.",
)
@extraflag
@pass_context
def build_cmd(
    ctx,
    *,
    output_path,
    watch,
    prune,
    verbosity,
    source_info_only,
    buildstate_path,
    extra_flags,
):
    """Builds the entire project into the final artifacts.

    The default behavior is to build the project into the default build
    output path which can be discovered with the `project-info` command
    but an alternative output folder can be provided with the `--output-path`
    option.

    The default behavior is to perform a build followed by a pruning step
    which removes no longer referenced artifacts from the output folder.
    Lektor will only build the files that require rebuilding if the output
    folder is reused.

    To enforce a clean build you have to issue a `clean` command first.

    If the build fails the exit code will be `1` otherwise `0`.  This can be
    used by external scripts to only deploy on successful build for instance.
    """
    from lektor.builder import Builder
    from lektor.reporter import CliReporter

    if output_path is None:
        output_path = ctx.get_default_output_path()

    ctx.load_plugins(extra_flags=extra_flags)

    env = ctx.get_env()

    with CliReporter(env, verbosity=verbosity):
        builds = ["first"]
        if watch:
            from lektor.watcher import watch_project

            click.secho("Watching for file system changes", fg="cyan")
            builds = chain(
                builds, watch_project(env, output_path, raise_interrupt=False)
            )

        success = False
        for _ in builds:
            builder = Builder(
                env.new_pad(),
                output_path,
                buildstate_path=buildstate_path,
                extra_flags=extra_flags,
            )
            if source_info_only:
                builder.update_all_source_infos()
                success = True
            else:
                failures = builder.build_all()
                if prune:
                    builder.prune()
                success = failures == 0

        return sys.exit(0 if success else 1)


@cli.command("clean")
@click.option(
    "-O",
    "--output-path",
    type=ResolvedPath(writable=True, file_okay=False),
    default=None,
    help="The output path.",
)
@click.option(
    "-v",
    "--verbose",
    "verbosity",
    count=True,
    help="Increases the verbosity of the logging.",
)
@click.confirmation_option(help="Confirms the cleaning.")
@extraflag
@pass_context
def clean_cmd(ctx, *, output_path, verbosity, extra_flags):
    """Cleans the entire build folder.

    If not build folder is provided, the default build folder of the project
    in the Lektor cache is used.
    """
    from lektor.builder import Builder
    from lektor.reporter import CliReporter

    if output_path is None:
        output_path = ctx.get_default_output_path()

    ctx.load_plugins(extra_flags=extra_flags)
    env = ctx.get_env()

    reporter = CliReporter(env, verbosity=verbosity)
    with reporter:
        builder = Builder(env.new_pad(), output_path)
        builder.prune(all=True)


@cli.command("deploy", short_help="Deploy the website.")
@click.argument("server", required=False)
@click.option(
    "-O",
    "--output-path",
    type=ResolvedPath(writable=True, file_okay=False),
    default=None,
    help="The output path.",
)
@click.option(
    "--username",
    envvar="LEKTOR_DEPLOY_USERNAME",
    help="An optional username to override the URL.",
)
@click.option(
    "--password",
    envvar="LEKTOR_DEPLOY_PASSWORD",
    help="An optional password to override the URL or the default prompt.",
)
@click.option(
    "--key-file",
    envvar="LEKTOR_DEPLOY_KEY_FILE",
    help="The path to a key file that should be used for the "
    "authentication of the deployment.",
)
@click.option(
    "--key",
    envvar="LEKTOR_DEPLOY_KEY",
    help="The contents of a key file directly a string that should "
    "be used for authentication of the deployment.",
)
@extraflag
@pass_context
def deploy_cmd(ctx, *, server, output_path, extra_flags, **credentials):
    """This command deploys the entire contents of the build folder
    (`--output-path`) onto a configured remote server.  The name of the
    server must fit the name from a target in the project configuration.
    If no server is supplied then the default server from the config is
    used.

    The deployment credentials are typically contained in the project config
    file but it's also possible to supply them here explicitly.  In this
    case the `--username` and `--password` parameters (as well as the
    `LEKTOR_DEPLOY_USERNAME` and `LEKTOR_DEPLOY_PASSWORD` environment
    variables) can override what's in the URL.

    For more information see the deployment chapter in the documentation.
    """
    from lektor.publisher import publish
    from lektor.publisher import PublishError

    if output_path is None:
        output_path = ctx.get_default_output_path()

    ctx.load_plugins(extra_flags=extra_flags)
    env = ctx.get_env()
    config = env.load_config()

    if server is None:
        server_info = config.get_default_server()
        if server_info is None:
            raise click.BadParameter(
                "No default server configured.", param_hint="server"
            )
    else:
        server_info = config.get_server(server)
        if server_info is None:
            raise click.BadParameter(
                f"Server {server!r} does not exist.", param_hint="server"
            )

    try:
        event_iter = publish(
            env,
            server_info.target,
            output_path,
            credentials=credentials,
            server_info=server_info,
            extra_flags=extra_flags,
        )
    except PublishError as exc:
        server_desc = "Default server" if server is None else f"Server {server!r}"
        message = f"{server_desc} configuration error: {exc}"
        raise click.UsageError(message) from exc

    click.echo(f"Deploying to {server_info.name}")
    click.echo(f"  Build cache: {output_path}")
    click.echo(f"  Target: {secure_url(server_info.target)}")
    try:
        for line in event_iter:
            click.echo(f"  {click.style(line, fg='cyan')}")
    except PublishError as e:
        click.secho(f"Error: {e}", fg="red")
    else:
        click.echo("Done!")


@cli.command("server", short_help="Launch a local server.")
@click.option(
    "-h",
    "--host",
    default="127.0.0.1",
    help="The network interface to bind to.  The default is the "
    "loopback device, but by setting it to 0.0.0.0 it becomes "
    "available on all network interfaces.",
)
@click.option(
    "-p", "--port", default=5000, help="The port to bind to.", show_default=True
)
@click.option(
    "-O",
    "--output-path",
    type=ResolvedPath(writable=True, file_okay=False),
    default=None,
    help="The dev server will build into the same folder as "
    "the build command by default.",
)
@pruneflag
@click.option(
    "-v",
    "--verbose",
    "verbosity",
    count=True,
    help="Increases the verbosity of the logging.",
)
@extraflag
@click.option("--browse", is_flag=True)
@pass_context
def server_cmd(ctx, *, host, port, output_path, prune, verbosity, extra_flags, browse):
    """The server command will launch a local server for development.

    Lektor's development server will automatically build all files into
    pages similar to how the build command with the `--watch` switch
    works, but also at the same time serve up the website on a local
    HTTP server.
    """
    from lektor.devserver import run_server

    if output_path is None:
        output_path = ctx.get_default_output_path()
    ctx.load_plugins(extra_flags=extra_flags)
    click.echo(f" * Project path: {ctx.get_project().project_path}")
    click.echo(f" * Output path: {output_path}")
    run_server(
        (host, port),
        env=ctx.get_env(),
        output_path=output_path,
        prune=prune,
        verbosity=verbosity,
        ui_lang=ctx.ui_lang,
        extra_flags=extra_flags,
        lektor_dev=os.environ.get("LEKTOR_DEV") == "1",
        browse=browse,
    )


@cli.command("project-info", short_help="Shows the info about a project.")
@click.option("as_json", "--json", is_flag=True, help="Prints out the data as json.")
@click.option(
    "--name",
    is_flag=True,
    help="Print the project name",
)
@click.option(
    "--project-file",
    is_flag=True,
    help="Print the path to the project file.",
)
@click.option(
    "--tree",
    is_flag=True,
    help="Print the path to the tree.",
)
@click.option(
    "default_output_path",
    "--output-path",
    is_flag=True,
    help="Print the path to the default output path.",
)
@click.option(
    "package_cache_path",
    "--package-cache",
    is_flag=True,
    help="Print the path to the package cache.",
)
@pass_context
def project_info_cmd(ctx, *, as_json, **opts):
    """Prints out information about the project.  This is particular
    useful for script usage or for discovering information about a
    Lektor project that is not immediately obvious (like the paths
    to the default output folder).
    """
    project = ctx.get_project()
    json_data = project.to_json()
    if as_json:
        echo_json(json_data)
        return

    ops = [k for k, v in opts.items() if v]
    if ops:
        for op in ops:
            click.echo(json_data.get(op, ""))
    else:
        click.echo(f"Name: {json_data['name']}")
        click.echo(f"File: {json_data['project_file']}")
        click.echo(f"Tree: {json_data['tree']}")
        click.echo(f"Output: {json_data['default_output_path']}")
        click.echo(f"Package Cache: {json_data['package_cache_path']}")


@cli.command(
    "content-file-info", short_help="Provides information for a set of lektor files."
)
@click.option("as_json", "--json", is_flag=True, help="Prints out the data as json.")
@click.argument("files", nargs=-1, type=click.Path(dir_okay=False))
@pass_context
def content_file_info_cmd(ctx, files, *, as_json):
    """Given a list of files this returns the information for those files
    in the context of a project.  If the files are from different projects
    an error is generated.
    """
    project = None

    def fail(msg):
        if as_json:
            echo_json({"success": False, "error": msg})
            sys.exit(1)
        raise click.UsageError(f"Could not find content file info: {msg}")

    for filename in files:
        this_project = Project.discover(filename)
        if this_project is None:
            fail("no project found")
        if project is None:
            project = this_project
        elif project.project_path != this_project.project_path:
            fail("multiple projects")

    if project is None:
        fail("no file indicated a project")

    project_files = []
    for filename in files:
        content_path = project.content_path_from_filename(filename)
        if content_path is not None:
            project_files.append(content_path)

    if not project_files:
        fail("no files resolve in project")

    if as_json:
        echo_json(
            {
                "success": True,
                "project": project.to_json(),
                "paths": project_files,
            }
        )
    else:
        click.echo("Project:")
        click.echo(f"  Name: {project.name}")
        click.echo(f"  File: {project.project_file}")
        click.echo(f"  Tree: {project.tree}")
        click.echo("Paths:")
        for project_file in project_files:
            click.echo(f"  - {project_file}")


@cli.group("plugins", short_help="Manages plugins.")
def plugins_cmd():
    """This command group provides various helpers to manages plugins
    in a Lektor project.
    """


@plugins_cmd.command("add", short_help="Adds a new plugin to the project.")
@click.argument("name")
@pass_context
def plugins_add_cmd(ctx, name):
    """This command can add a new plugin to the project.  If just given
    the name of the plugin the latest version of that plugin is added to
    the project.

    The argument is either the name of the plugin or the name of the plugin
    suffixed with `@version` with the version.  For instance to install
    the version 0.1 of the plugin demo you would do `demo@0.1`.
    """
    project = ctx.get_project()
    from .packages import add_package_to_project

    try:
        info = add_package_to_project(project, name)
    except RuntimeError as e:
        click.echo(f"Error: {e}", err=True)
    else:
        click.echo(
            f"Package {info['name']} ({info['version']}) was added to the project"
        )


@plugins_cmd.command("remove", short_help="Removes a plugin from the project.")
@click.argument("name")
@pass_context
def plugins_remove_cmd(ctx, name):
    """This command can remove a plugin to the project again.  The name
    of the plugin is the only argument to the function.
    """
    project = ctx.get_project()
    from .packages import remove_package_from_project

    try:
        old_info = remove_package_from_project(project, name)
    except RuntimeError as e:
        click.echo(f"Error: {e}", err=True)
    else:
        if old_info is None:
            click.echo(
                "Package was not registered with the project. Nothing was removed."
            )
        else:
            click.echo(f"Removed package {old_info['name']} ({old_info['version']})")


@plugins_cmd.command("list", short_help="List all plugins.")
@click.option("as_json", "--json", is_flag=True, help="Prints out the data as json.")
@click.option(
    "-v",
    "--verbose",
    "verbosity",
    count=True,
    help="Increases the verbosity of the output.",
)
@pass_context
def plugins_list_cmd(ctx, *, as_json, verbosity):
    """This returns a list of all currently actively installed plugins
    in the project.  By default it only prints out the plugin IDs and
    version numbers but the entire information can be returned by
    increasing verbosity with `-v`.  Additionally JSON output can be
    requested with `--json`.
    """
    ctx.load_plugins()
    env = ctx.get_env()
    plugins = sorted(env.plugins.values(), key=lambda x: x.id.lower())

    if as_json:
        echo_json({"plugins": [x.to_json() for x in plugins]})
        return

    if verbosity == 0:
        for plugin in plugins:
            click.echo(f"{plugin.id} (version {plugin.version})")
        return

    for idx, plugin in enumerate(plugins):
        if idx:
            click.echo()
        click.echo(f"{plugin.name} ({plugin.id})")
        for line in plugin.description.splitlines():
            click.echo(f"  {line}")
        if plugin.path is not None:
            click.echo(f"  path: {plugin.path}")
        click.echo(f"  version: {plugin.version}")
        click.echo(f"  import-name: {plugin.import_name}")


@plugins_cmd.command("flush-cache", short_help="Flushes the plugin installation cache.")
@pass_context
def plugins_flush_cache_cmd(ctx):
    """This uninstalls all plugins in the cache.  On next usage the plugins
    will be reinstalled automatically.  This is mostly just useful during
    plugin development when the cache got corrupted.
    """
    click.echo("Flushing plugin cache ...")
    from .packages import wipe_package_cache

    wipe_package_cache(ctx.get_env())
    click.echo("All done!")


@plugins_cmd.command("reinstall", short_help="Reinstall all plugins.")
@pass_context
def plugins_reinstall_cmd(ctx):
    """Forces a re-installation of all plugins.  This will download the
    requested versions of the plugins and install them into the plugin
    cache folder.  Alternatively one can just use `flush-cache` to
    flush the cache and on next build Lektor will automatically download
    the plugins again.
    """
    ctx.load_plugins(reinstall=True)


@cli.command("quickstart", short_help="Starts a new empty project.")
@click.option("--name", help="The name of the project.")
@click.option(
    "--path",
    type=click.Path(file_okay=False, dir_okay=False, writable=True),
    help="Output directory",
)
@pass_context
def quickstart_cmd(ctx, **options):
    """Starts a new empty project with a minimum boilerplate."""
    from lektor.quickstart import project_quickstart

    project_quickstart(options)


cli.add_command(devcli, "dev")


def main(as_module=False):
    args = sys.argv[1:]
    name = None

    if as_module:
        name = "python -m lektor"
        sys.argv = ["-m", "lektor"] + args

    cli.main(args=args, prog_name=name)


if __name__ == "__main__":
    main(as_module=True)



import os
import stat
import warnings
from contextlib import contextmanager
from dataclasses import dataclass
from itertools import islice
from pathlib import Path
from urllib.parse import urlsplit

import pytest

from lektor.utils import atomic_open
from lektor.utils import build_url
from lektor.utils import create_temp
from lektor.utils import deprecated
from lektor.utils import is_path_child_of
from lektor.utils import join_path
from lektor.utils import magic_split_ext
from lektor.utils import make_relative_url
from lektor.utils import parse_path
from lektor.utils import secure_url
from lektor.utils import slugify
from lektor.utils import unique_everseen
from lektor.utils import untrusted_to_os_path
from lektor.utils import Url


@pytest.mark.parametrize(
    "head, tail, expected",
    [
        ("a", "b", "a/b"),
        ("/a", "b", "/a/b"),
        ("a@b", "c", "a@b/c"),
        ("a@b", "", "a@b"),
        ("a@b", "@c", "a@c"),
        ("a@b/c", "a@b", "a/a@b"),
        #
        ("/a", "/", "/"),
        ("/a", "/b", "/b"),
        ("a@b", "/", "/"),
        ("a@b", "/c", "/c"),
        ("a@b", "/c@d", "/c@d"),
        #
        ("blog@archive", "2015", "blog@archive/2015"),
        ("blog@archive/2015", "..", "blog@archive"),
        ("blog@archive/2015", "@archive", "blog@archive"),
        ("blog@archive", "..", "blog"),
        ("blog@archive", ".", "blog@archive"),
        ("blog@archive", "", "blog@archive"),
        #
        # special behavior: parent of pagination paths is always the actual
        # page parent.
        ("/blog@1", "..", "/"),
        ("/blog@2", "..", "/"),
        # But joins on the same level keep the path
        ("/blog@1", ".", "/blog@1"),
        ("/blog@2", ".", "/blog@2"),
        ("/blog@1", "", "/blog@1"),
        ("/blog@2", "", "/blog@2"),
    ],
)
def test_join_path(head, tail, expected):
    assert join_path(head, tail) == expected


def test_is_path_child_of():
    assert not is_path_child_of("a/b", "a/b")
    assert is_path_child_of("a/b", "a/b", strict=False)
    assert is_path_child_of("a/b/c", "a")
    assert not is_path_child_of("a/b/c", "b")
    assert is_path_child_of("a/b@foo/bar", "a/b@foo")
    assert is_path_child_of("a/b@foo", "a/b@foo", strict=False)
    assert not is_path_child_of("a/b@foo/bar", "a/c@foo")
    assert not is_path_child_of("a/b@foo/bar", "a/c")
    assert is_path_child_of("a/b@foo", "a/b")
    assert is_path_child_of("a/b@foo/bar", "a/b@foo")
    assert not is_path_child_of("a/b@foo/bar", "a/b@bar")


def test_magic_split_ext():
    assert magic_split_ext("wow") == ("wow", "")
    assert magic_split_ext("aaa.jpg") == ("aaa", "jpg")
    assert magic_split_ext("aaa. jpg") == ("aaa. jpg", "")
    assert magic_split_ext("aaa.j pg") == ("aaa.j pg", "")
    assert magic_split_ext("aaa.j pg", ext_check=False) == ("aaa", "j pg")


def test_slugify():
    assert slugify("w o w") == "w-o-w"
    assert slugify("Șö prĕtty") == "so-pretty"
    assert slugify("im age.jpg") == "im-age.jpg"
    assert slugify("slashed/slug") == "slashed/slug"


def test_Url_constructor_deprecated():
    with pytest.deprecated_call():
        Url("https://example.org")


@dataclass
class SampleUrl:
    uri: str
    iri: str

    @property
    def split_uri(self):
        return urlsplit(self.uri)

    @property
    def split_iri(self):
        return urlsplit(self.iri)


SAMPLE_URLS = [
    SampleUrl("https://example.org/foo", "https://example.org/foo"),
    SampleUrl("https://example.org:8001/f%C3%BC", "https://example.org:8001/fü"),
    SampleUrl(
        "https://xn--wgv71a119e.idn.icann.org/%E5%A4%A7",
        "https://日本語.idn.icann.org/大",
    ),
    SampleUrl("/?q=sch%C3%B6n#gru%C3%9F", "/?q=schön#gruß"),
]


@pytest.fixture(params=SAMPLE_URLS, ids=lambda sample: sample.uri)
def sample_url(request):
    sample_url = request.param
    # sanity checks
    assert sample_url.split_uri.scheme == sample_url.split_iri.scheme
    assert sample_url.split_uri.port == sample_url.split_iri.port
    return sample_url


def test_Url_str(sample_url):
    assert str(Url.from_string(sample_url.iri)) == sample_url.iri
    assert str(Url.from_string(sample_url.uri)) == sample_url.uri


def test_Url_ascii_url(sample_url):
    assert Url.from_string(sample_url.iri).ascii_url == sample_url.uri
    assert Url.from_string(sample_url.uri).ascii_url == sample_url.uri


def test_Url_ascii_host(sample_url):
    assert Url.from_string(sample_url.iri).ascii_host == sample_url.split_uri.hostname
    assert Url.from_string(sample_url.uri).ascii_host == sample_url.split_uri.hostname


def test_Url_scheme(sample_url):
    assert Url.from_string(sample_url.iri).scheme == sample_url.split_uri.scheme
    assert Url.from_string(sample_url.uri).scheme == sample_url.split_uri.scheme


def test_Url_host(sample_url):
    assert Url.from_string(sample_url.iri).host == sample_url.split_iri.hostname
    assert Url.from_string(sample_url.uri).host == sample_url.split_iri.hostname


def test_Url_port(sample_url):
    assert Url.from_string(sample_url.iri).port == sample_url.split_uri.port
    assert Url.from_string(sample_url.uri).port == sample_url.split_uri.port


def test_Url_path(sample_url):
    assert Url.from_string(sample_url.iri).path == sample_url.split_iri.path
    assert Url.from_string(sample_url.uri).path == sample_url.split_iri.path


def test_Url_query(sample_url):
    assert Url.from_string(sample_url.iri).query == sample_url.split_iri.query
    assert Url.from_string(sample_url.uri).query == sample_url.split_iri.query


def test_Url_anchor(sample_url):
    assert Url.from_string(sample_url.iri).anchor == sample_url.split_iri.fragment
    assert Url.from_string(sample_url.uri).anchor == sample_url.split_iri.fragment


@pytest.mark.parametrize(
    "url, expected",
    [
        ("https://user:pw@example.org/p", "https://user@example.org/p"),
        ("https://user:pw@example.org:8000", "https://user@example.org:8000"),
        ("https://user@example.org/b", "https://user@example.org/b"),
    ],
)
def test_secure_url(url, expected):
    assert secure_url(url) == expected


def test_url_builder():
    assert build_url([]) == "/"
    assert build_url(["a", "b/c"]) == "/a/b/c/"
    assert build_url(["a", "b/c"], trailing_slash=False) == "/a/b/c"
    assert build_url(["a", "b/c.html"]) == "/a/b/c.html"
    assert build_url(["a", "b/c.html"], trailing_slash=True) == "/a/b/c.html/"
    assert build_url(["a", None, "b", "", "c"]) == "/a/b/c/"


def test_url_builder_dot_handling():
    assert build_url(["blog", "01.01.2025", "photos"]) == "/blog/01.01.2025/photos/"


def test_parse_path():
    assert parse_path("") == []
    assert parse_path("/") == []
    assert parse_path("/foo") == ["foo"]
    assert parse_path("/foo/") == ["foo"]
    assert parse_path("/foo/bar") == ["foo", "bar"]
    assert parse_path("/foo/bar/") == ["foo", "bar"]
    assert parse_path("/foo/bar/../stuff") == ["foo", "stuff"]


@pytest.mark.parametrize(
    "source, target, expected",
    [
        ("/", "./a/", "a/"),
        ("/", "./a", "a"),
        ("/fr/blog/2015/11/a/", "/fr/blog/2015/11/a/a.jpg", "a.jpg"),
        ("/fr/blog/2015/11/a/", "/fr/blog/", "../../../"),
        ("/fr/blog/2015/11/a.php", "/fr/blog/", "../../"),
        ("/fr/blog/2015/11/a/", "/fr/blog/2016/", "../../../2016/"),
        ("/fr/blog/2015/11/a/", "/fr/blog/2016/c.jpg", "../../../2016/c.jpg"),
        ("/fr/blog/2016/", "/fr/blog/2015/a/", "../2015/a/"),
        ("/fr/blog/2016/", "/fr/blog/2015/a/d.jpg", "../2015/a/d.jpg"),
        ("/fr/blog/2015/11/a/", "/images/b.svg", "../../../../../images/b.svg"),
        ("/fr/blog/", "2015/11/", "2015/11/"),
        ("/fr/blog/x", "2015/11/", "2015/11/"),
        ("", "./a/", "a/"),
        ("", "./a", "a"),
        ("fr/blog/2015/11/a/", "fr/blog/2015/11/a/a.jpg", "a.jpg"),
        ("fr/blog/2015/11/a/", "fr/blog/", "../../../"),
        ("fr/blog/2015/11/a.php", "fr/blog/", "../../"),
        ("fr/blog/2015/11/a/", "fr/blog/2016/", "../../../2016/"),
        ("fr/blog/2015/11/a/", "fr/blog/2016/c.jpg", "../../../2016/c.jpg"),
        ("fr/blog/2016/", "fr/blog/2015/a/", "../2015/a/"),
        ("fr/blog/2016/", "fr/blog/2015/a/d.jpg", "../2015/a/d.jpg"),
        ("fr/blog/2015/11/a/", "images/b.svg", "../../../../../images/b.svg"),
        ("fr/blog/", "2015/11/", "../../2015/11/"),
        ("fr/blog/x", "2015/11/", "../../2015/11/"),
    ],
)
def test_make_relative_url(source, target, expected):
    assert make_relative_url(source, target) == expected


def test_make_relative_url_relative_source_absolute_target():
    with pytest.raises(ValueError):
        make_relative_url("rel/a/tive/", "/abs/o/lute")


def test_atomic_open(tmp_path):
    path = tmp_path / "test.txt"
    path.write_text("previous")

    with atomic_open(path, "w") as fp:
        fp.write("new")
        fp.flush()
        assert path.read_text() == "previous"
        assert len(list(tmp_path.iterdir())) == 2
    assert path.read_text() == "new"
    assert len(list(tmp_path.iterdir())) == 1


def test_atomic_open_exception(tmp_path):
    path = tmp_path / "test.txt"

    with pytest.raises(RuntimeError, match="test"):
        with atomic_open(path, "w") as fp:
            fp.write("new")
            fp.flush()
            raise RuntimeError("test")
    assert len(list(tmp_path.iterdir())) == 0


@pytest.fixture(params=[0o022, 0o002, 0x000])
def umask(request):
    umask = request.param
    if os.name == "nt":
        pytest.skip("Windows does not support umask")
    saved = os.umask(umask)
    try:
        yield umask
    finally:
        os.umask(saved)


def test_atomic_open_respects_umask(tmp_path, umask):
    path = tmp_path / "test.txt"
    with atomic_open(path, "w"):
        pass
    assert oct(stat.S_IMODE(path.stat().st_mode)) == oct(0o666 & ~umask)


@pytest.mark.parametrize("mode", ["a", "w+", "x", "rw", "foo"])
def test_atomic_open_raises_on_bad_mode(tmp_path, mode):
    with pytest.raises(ValueError, match="mode"):
        with atomic_open(tmp_path / "file.txt", mode):
            pass


def test_create_temp(tmp_path):
    assert sum(1 for _ in tmp_path.iterdir()) == 0
    fd, filename = create_temp(prefix="a", suffix=".bin", dir=tmp_path)
    assert sum(1 for _ in tmp_path.iterdir()) == 1
    assert Path(filename).parent == tmp_path
    os.write(fd, b"test\n")
    os.close(fd)
    assert Path(filename).read_bytes() == b"test\n"


@pytest.mark.parametrize("mode", [0o666, 0o777])
def test_create_temp_respects_umask(tmp_path, mode, umask):
    _, filename = create_temp(dir=tmp_path, mode=mode)
    assert oct(stat.S_IMODE(os.stat(filename).st_mode)) == oct(mode & ~umask)


@pytest.mark.parametrize(
    "seq, expected",
    [
        (iter(()), ()),
        ((2, 1, 1, 2, 1), (2, 1)),
        ((1, 2, 1, 2, 1), (1, 2)),
    ],
)
def test_unique_everseen(seq, expected):
    assert tuple(unique_everseen(seq)) == expected


@contextmanager
def _local_deprecated_call(match=None):
    """Like pytest.deprecated_call, but also check that all warnings
    are attributed to this file.
    """
    with pytest.deprecated_call(match=match) as warnings:
        yield warnings
    assert all(w.filename == __file__ for w in warnings)


@pytest.mark.parametrize(
    "kwargs, match",
    [
        ({}, r"^'f' is deprecated$"),
        ({"reason": "testing"}, r"^'f' is deprecated \(testing\)$"),
        (
            {"reason": "testing", "version": "1.2.3"},
            r"^'f' is deprecated \(testing\) since version 1.2.3$",
        ),
        ({"name": "oldfunc"}, r"^'oldfunc' is deprecated$"),
    ],
)
def test_deprecated_function(kwargs, match):
    @deprecated(**kwargs)
    def f():
        return 42

    with _local_deprecated_call(match=match):
        assert f() == 42


def test_deprecated_method():
    class Cls:
        @deprecated
        def f(self):  # pylint: disable=no-self-use
            return 42

        @deprecated
        def g(self):
            return self.f()

    with _local_deprecated_call(match=r"^'g' is deprecated$") as warnings:
        assert Cls().g() == 42
    assert len(warnings) == 1


def test_deprecated_classmethod():
    class Cls:
        @classmethod
        @deprecated
        def f(cls):
            return 42

        @classmethod
        @deprecated
        def g(cls):
            return cls.f()

    with _local_deprecated_call(match=r"^'g' is deprecated$") as warnings:
        assert Cls().g() == 42
    assert len([w.message for w in warnings]) == 1


def test_deprecated_raises_type_error():
    with pytest.raises(TypeError):
        deprecated(0)


def test_deprecated_stacklevel():
    @deprecated(stacklevel=2)
    def f():
        return 42

    def g():
        return f()

    with _local_deprecated_call(match=r"^'f' is deprecated$") as warnings:
        assert g() == 42
    assert "assert g() == 42" in _warning_line(warnings[0])


def _warning_line(warning: warnings.WarningMessage) -> str:
    """Get the text of the line for which warning was issued."""
    with open(warning.filename, encoding="utf-8") as fp:
        return next(islice(fp, warning.lineno - 1, None), None)


@pytest.mark.parametrize(
    "db_path, expected",
    [
        ("a/b", "a/b"),
        ("/a/b", "a/b"),
        ("a/b/", "a/b"),
        ("/../../a", "a"),
    ],
)
def test_untrusted_to_os_path(db_path, expected):
    os_path = untrusted_to_os_path(db_path)
    assert os_path.split(os.sep) == expected.split("/")



from __future__ import annotations

import inspect
import os
import sys
import warnings
from importlib import metadata
from pathlib import Path
from weakref import ref as weakref

from inifile import IniFile

from lektor.context import get_ctx
from lektor.utils import process_extra_flags


def get_plugin(plugin_id_or_class, env=None):
    """Looks up the plugin instance by id or class."""
    if env is None:
        ctx = get_ctx()
        if ctx is None:
            raise RuntimeError(
                "Context is unavailable and no environment was passed to the function."
            )
        env = ctx.env
    plugin_id = env.plugin_ids_by_class.get(plugin_id_or_class, plugin_id_or_class)
    try:
        return env.plugins[plugin_id]
    except KeyError as error:
        raise LookupError(f"Plugin {plugin_id!r} not found") from error


class Plugin:
    """This needs to be subclassed for custom plugins."""

    name = "Your Plugin Name"
    description = "Description goes here"

    __dist: metadata.Distribution = None

    def __init__(self, env, id):
        self._env = weakref(env)
        self.id = id

    @property
    def env(self):
        rv = self._env()
        if rv is None:
            raise RuntimeError("Environment went away")
        return rv

    @property
    def version(self):
        if self.__dist is not None:
            return self.__dist.version
        return None

    @property
    def path(self) -> str:
        mod = sys.modules[self.__class__.__module__.split(".", maxsplit=1)[0]]
        path = Path(mod.__file__).resolve().parent
        package_cache = self.env.project.get_package_cache_path()
        try:
            # We could use Path.is_relative_to(), except that's py39+ only
            path.relative_to(package_cache)
            # We're only interested in local, editable packages. This is not one.
            return None
        except ValueError:
            pass
        return os.fspath(path)

    @property
    def import_name(self):
        return self.__class__.__module__ + ":" + self.__class__.__name__

    def get_lektor_config(self):
        """Returns the global config."""
        ctx = get_ctx()
        if ctx is not None:
            cfg = ctx.pad.db.config
        else:
            cfg = self.env.load_config()
        return cfg

    @property
    def config_filename(self):
        """The filename of the plugin specific config file."""
        return os.path.join(self.env.root_path, "configs", self.id + ".ini")

    def get_config(self, fresh=False):
        """Returns the config specific for this plugin.  By default this
        will be cached for the current build context but this can be
        disabled by passing ``fresh=True``.
        """
        ctx = get_ctx()
        if ctx is not None and not fresh:
            cache = ctx.cache.setdefault(__name__ + ":configs", {})
            cfg = cache.get(self.id)
            if cfg is None:
                cfg = IniFile(self.config_filename)
                cache[self.id] = cfg
        else:
            cfg = IniFile(self.config_filename)
        if ctx is not None:
            ctx.record_dependency(self.config_filename)
        return cfg

    def emit(self, event, **kwargs):
        return self.env.plugin_controller.emit(self.id + "-" + event, **kwargs)

    def to_json(self):
        return {
            "id": self.id,
            "name": self.name,
            "version": self.version,
            "description": self.description,
            "path": self.path,
            "import_name": self.import_name,
        }


def _find_plugins():
    """Find all available plugins.

    Returns an interator of (distribution, entry_point) pairs.
    """
    for dist in metadata.distributions():
        for ep in dist.entry_points:
            if ep.group == "lektor.plugins":
                _check_dist_name(dist.metadata["Name"], ep.name)
                yield dist, ep


def _check_dist_name(dist_name, plugin_id):
    """Check that plugin comes from a validly named distribution.

    Raises RuntimeError if distribution name is not of the form
    ``lektor-``*<plugin_id>*.
    """
    # XXX: do we really need to be so strict about distribution names?
    match_name = "lektor-" + plugin_id.lower()
    if match_name != dist_name.lower():
        raise RuntimeError(
            "Disallowed distribution name: distribution name for "
            f"plugin {plugin_id!r} must be {match_name!r} (not {dist_name!r})."
        )


def initialize_plugins(env):
    """Initializes the plugins for the environment."""
    for dist, ep in _find_plugins():
        plugin_id = ep.name
        plugin_cls = ep.load()
        env.plugin_controller.instanciate_plugin(plugin_id, plugin_cls, dist)
    env.plugin_controller.emit("setup-env")


class PluginController:
    """Helper management class that is used to control plugins through
    the environment.
    """

    def __init__(self, env, extra_flags=None):
        self._env = weakref(env)
        self.extra_flags = extra_flags

    @property
    def env(self):
        rv = self._env()
        if rv is None:
            raise RuntimeError("Environment went away")
        return rv

    def instanciate_plugin(
        self,
        plugin_id: str,
        plugin_cls: type[Plugin],
        dist: metadata.Distribution | None = None,
    ) -> None:
        env = self.env
        if plugin_id in env.plugins:
            raise RuntimeError(f'Plugin "{plugin_id}" is already registered')
        plugin = plugin_cls(env, plugin_id)
        # Plugin.version needs the source distribution to be able to cleanly determine
        # the plugin version.  For reasons of backward compatibility, we don't want to
        # change the signature of the constructor, so we stick it in a private attribute
        # here.
        plugin._Plugin__dist = dist
        env.plugins[plugin_id] = plugin
        env.plugin_ids_by_class[plugin_cls] = plugin_id

    def iter_plugins(self):
        # XXX: sort?
        return self.env.plugins.values()

    def emit(self, event, **kwargs):
        """Invoke event hook for all plugins that support it.

        Any ``kwargs`` are passed to the hook methods.

        Returns a dict mapping plugin ids to hook method return values.
        """
        rv = {}
        extra_flags = process_extra_flags(self.extra_flags)
        funcname = "on_" + event.replace("-", "_")
        for plugin in self.iter_plugins():
            handler = getattr(plugin, funcname, None)
            if handler is not None:
                kw = {**kwargs, "extra_flags": extra_flags}
                try:
                    inspect.signature(handler).bind(**kw)
                except TypeError:
                    del kw["extra_flags"]
                rv[plugin.id] = handler(**kw)
                if "extra_flags" not in kw:
                    warnings.warn(
                        # deprecated since 3.2.0
                        f"The plugin {plugin.id!r} function {funcname!r} does not "
                        "accept extra_flags. "
                        "It should be updated to accept `**extra` so that it will "
                        "not break if new parameters are passed to it by newer "
                        "versions of Lektor.",
                        DeprecationWarning,
                        stacklevel=2,
                    )
        return rv



import re

import pytest

from lektor.datamodel import ChildConfig
from lektor.datamodel import DataModel
from lektor.datamodel import PaginationConfig
from lektor.reporter import BufferReporter


@pytest.fixture
def slug_format():
    return None


@pytest.fixture
def datamodel(env, slug_format):
    id_ = "test"
    name_i18n = {"en": "test"}
    child_config = ChildConfig(slug_format=slug_format)
    return DataModel(env, id_, name_i18n, child_config=child_config)


@pytest.mark.parametrize(
    "per_page, expected",
    [
        (-1, ValueError),
        (-1.5, TypeError),
        ("5", TypeError),
    ],
)
def test_PaginationConfig_init_raises_on_invalid_per_page(env, per_page, expected):
    with pytest.raises(expected):
        PaginationConfig(env, per_page=per_page)


@pytest.mark.parametrize(
    "slug_format, expected_slug",
    [
        (None, "child-id"),
        ("fmt-{{ this._id }}", "fmt-child-id"),
        ("{{ unknown.attr }}", "temp-child-id"),
    ],
)
def test_get_default_child_slug(datamodel, pad, expected_slug):
    data = {
        "_id": "child-id",
    }
    assert datamodel.get_default_child_slug(pad, data) == expected_slug


@pytest.mark.parametrize("slug_format", ["{{ unknown.attr }}"])
def test_get_default_child_slug_reports_failure(datamodel, pad):
    data = {"_id": "id"}
    with BufferReporter(pad.env) as reporter:
        assert datamodel.get_default_child_slug(pad, data) == "temp-id"
    _, event_data = reporter.buffer[-1]
    message = event_data["message"]
    assert re.search(r"failed to expand\b.*\bslug_format\b", message, re.IGNORECASE)



from collections.abc import Callable
from collections.abc import Iterable
from collections.abc import Iterator
from functools import update_wrapper
from itertools import chain
from typing import Any

from flask import json
from flask import Response


def eventstream(f: Callable[..., Iterable[Any]]) -> Callable[..., Response]:
    def new_func(*args: Any, **kwargs: Any) -> Response:
        def generate() -> Iterator[bytes]:
            for event in chain(f(*args, **kwargs), (None,)):
                yield f"data: {json.dumps(event)}\n\n".encode()

        return Response(
            generate(), mimetype="text/event-stream", direct_passthrough=True
        )

    return update_wrapper(new_func, f)



from __future__ import annotations

import posixpath
import warnings
from collections import defaultdict
from collections.abc import Generator
from collections.abc import Iterable
from collections.abc import Sequence
from contextlib import suppress
from itertools import takewhile
from operator import methodcaller
from pathlib import Path
from typing import TYPE_CHECKING

from werkzeug.utils import cached_property

from lektor.sourceobj import SourceObject
from lektor.utils import deprecated
from lektor.utils import DeprecatedWarning


if TYPE_CHECKING:
    from _typeshed import StrPath

    from lektor.db import Pad


def get_asset_root(pad: Pad, asset_roots: Iterable[StrPath]) -> Directory:
    """Get the merged asset root.

    This represents a logical merging or overlaying of (possibly) multiple asset trees,
    rooted at directories given by ``asset_roots``.

    Any paths listed in ``asset_roots`` that do not refer to a directory
    are silently ignored.
    """
    root_paths = tuple(
        Path(root).absolute() for root in asset_roots if Path(root).is_dir()
    )
    return Directory(pad, parent=None, name="", paths=root_paths)


@deprecated(version="3.4.0")
def get_asset(pad: Pad, filename: str, parent: Asset | None = None) -> Asset | None:
    if parent is None:
        parent = pad.asset_root
    else:
        assert pad is parent.pad
    return parent.get_child(filename)


_FROM_URL_DEPRECATED = DeprecatedWarning(
    "from_url",
    reason="The `from_url` parameter of `Asset.get_child` is now ignored.",
    version="3.4.0",
)


class Asset(SourceObject):
    source_classification = "asset"
    artifact_extension = ""

    def __init__(
        self, pad: Pad, name: str, parent: Asset | None, paths: tuple[Path, ...]
    ):
        super().__init__(pad)
        self.name = name
        self.parent = parent
        self._paths = paths

    def iter_source_filenames(self) -> Generator[str]:
        yield from map(str, self._paths)

    @property
    def url_name(self) -> str:
        name = self.name
        base, ext = posixpath.splitext(name)

        # If this is a known extension from an attachment then convert it
        # to lowercase
        if ext.lower() in self.pad.db.config["ATTACHMENT_TYPES"]:
            ext = ext.lower()

        return base + ext + self.artifact_extension

    @property
    def url_path(self) -> str:
        if self.parent is None:
            return "/" + self.name
        return posixpath.join(self.parent.url_path, self.url_name)

    @property
    def artifact_name(self) -> str:
        if self.parent is not None:
            return self.parent.artifact_name.rstrip("/") + "/" + self.url_name
        return self.url_path

    @property
    def children(self) -> Iterable[Asset]:
        return ()

    # pylint: disable-next=no-self-use,useless-return
    def get_child(self, name: str, from_url: bool = False) -> Asset | None:
        if from_url:
            warnings.warn(_FROM_URL_DEPRECATED, stacklevel=2)
        return None

    def resolve_url_path(self, url_path: Sequence[str]) -> Asset | None:
        if len(url_path) == 0:
            return self
        return None

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} {self.artifact_name!r}>"


class Directory(Asset):
    """Represents a merged set of asset directories."""

    @property
    def children(self) -> Iterable[Asset]:
        return self._children_by_name.values()

    def get_child(self, name: str, from_url: bool = False) -> Asset | None:
        if from_url:
            warnings.warn(_FROM_URL_DEPRECATED, stacklevel=2)
        return self._children_by_name.get(name)

    @cached_property
    def _children_by_name(self) -> dict[str, Asset]:
        return {asset.name: asset for asset in self._iter_children()}

    def _iter_children(self) -> Generator[Asset]:
        env = self.pad.env
        candidates_by_name = defaultdict(list)
        for path in self._paths:
            with suppress(OSError):
                for child in path.iterdir():
                    if not env.is_uninteresting_source_name(child.name):
                        candidates_by_name[child.name].append(child)

        for name, candidates in candidates_by_name.items():
            leading_dirs = tuple(takewhile(methodcaller("is_dir"), candidates))
            if leading_dirs:
                # Merge directories at the top of the overlay stack.
                #
                # Directories overlayed above a non-directory shadow (hide) that
                # non-directory.  (That non-directory, in turn, shadows anything under
                # it.)
                yield Directory(self.pad, parent=self, name=name, paths=leading_dirs)
            else:
                # If first candidate is not a directory, it shadows any below it.
                path = candidates[0]
                asset_class = env.special_file_assets.get(path.suffix, File)
                yield asset_class(self.pad, parent=self, name=name, paths=(path,))

    @property
    def url_name(self) -> str:
        return self.name + self.artifact_extension

    @property
    def url_path(self) -> str:
        path = super().url_path
        if not path.endswith("/"):
            path += "/"
        return path

    def resolve_url_path(self, url_path: Sequence[str]) -> Asset | None:
        if len(url_path) == 0:
            return self

        # Optimization: try common case where file extension has not been mangled
        child = self.get_child(url_path[0])
        if child is not None:
            return child.resolve_url_path(url_path[1:])

        if len(url_path) == 1:
            # There are a number of ways a file extension can be mangled to form the
            # url_path. (See `Asset.url_name`.) It can be lower-cased, and/or it could
            # have had `artifact_extension` appended.  It's not easy to check all these
            # cases individually, so we'll just look for a matching child.
            for child in self.children:
                if child.url_name == url_path[0]:
                    return child
        return None

    @property
    def url_content_path(self) -> str:
        return self.url_path


class File(Asset):
    """Represents a static asset file."""



"""Compatibility with various versions of Pillow."""

from __future__ import annotations

from collections.abc import Iterable
from collections.abc import Mapping
from enum import IntEnum
from types import ModuleType
from types import SimpleNamespace

import PIL.ExifTags
import PIL.Image


__all__ = ["ExifTags", "Transpose", "UnidentifiedImageError"]

PILLOW_VERSION_INFO = tuple(map(int, PIL.__version__.split(".")))

if PILLOW_VERSION_INFO >= (9, 4):
    ExifTags: ModuleType | SimpleNamespace = PIL.ExifTags
else:
    # Pillow < 9.4 does not provide the PIL.ExifTags.{Base,GPS,IFD} enums. Here we
    # provide and ExifTags namespace which has them.

    def _reverse_map(mapping: Mapping[int, str]) -> dict[str, int]:
        return dict(map(reversed, mapping.items()))  # type: ignore[arg-type]

    ExifTags = SimpleNamespace(
        Base=IntEnum("Base", _reverse_map(PIL.ExifTags.TAGS)),
        GPS=IntEnum("GPS", _reverse_map(PIL.ExifTags.GPSTAGS)),
        IFD=IntEnum("IFD", [("Exif", 34665), ("GPSInfo", 34853)]),
        TAGS=PIL.ExifTags.TAGS,
        GPSTAGS=PIL.ExifTags.GPSTAGS,
    )


if hasattr(PIL.Image, "Transpose"):
    # pillow >= 9.1
    Transpose = PIL.Image.Transpose
else:

    def _make_enum(name: str, members: Iterable[str]) -> IntEnum:
        items = ((member, getattr(PIL.Image, member)) for member in members)
        return IntEnum(name, items)

    Transpose = _make_enum(  # type: ignore[misc, assignment]
        "Transpose",
        (
            "FLIP_LEFT_RIGHT",
            "FLIP_TOP_BOTTOM",
            "ROTATE_90",
            "ROTATE_180",
            "ROTATE_270",
            "TRANSPOSE",
            "TRANSVERSE",
        ),
    )


# UnidentifiedImageError only exists in Pillow >= 7.0.0
UnidentifiedImageError = getattr(PIL, "UnidentifiedImageError", OSError)



from typing import Any

from flask import Blueprint
from flask import current_app
from flask import render_template
from flask import request


bp = Blueprint("dash", __name__, url_prefix="/admin")


@bp.route("/", defaults={"page": ""})
@bp.route("/<any(edit, delete, preview, 'add-child', upload):page>", endpoint="app")
def app_view(**kwargs: Any) -> str:
    """Render the React admin GUI app."""
    # Note: client side app handles redirect from page='' to page='edit'
    return render_template(
        "dash.html",
        lektor_config={
            "admin_root": request.root_path,
            "site_root": request.environ["lektor.site_root"],
            "lang": current_app.config.get("lektor.ui_lang", "en"),
        },
    )



import os
from inspect import cleandoc
from io import BytesIO
from operator import itemgetter
from pathlib import Path
from urllib.parse import urlencode

import pytest

from lektor.admin import WebAdmin
from lektor.admin.context import LektorContext
from lektor.admin.utils import eventstream
from lektor.builder import Builder
from lektor.constants import PRIMARY_ALT
from lektor.db import Database
from lektor.editor import EditorSession
from lektor.environment import Environment
from lektor.project import Project
from lektor.publisher import PublishError


def write_text(path, text):
    path.parent.mkdir(exist_ok=True)
    path.write_text(cleandoc(text))


@pytest.fixture
def scratch_project_data(scratch_project_data):
    write_text(
        scratch_project_data / "content/page/contents.lr",
        """
        _model: page
        ---
        title: A Page
        """,
    )
    return scratch_project_data


@pytest.fixture
def scratch_client(scratch_env, scratch_project):
    webadmin = WebAdmin(scratch_env, output_path=scratch_project.tree)
    with webadmin.test_client() as client:
        yield client


@pytest.fixture
def scratch_content_path(scratch_project):
    return Path(scratch_project.tree) / "content"


@pytest.fixture(scope="session")
def project_path(data_path):
    return data_path / "demo-project"


@pytest.fixture(scope="session")
def webadmin(tmp_path_factory, project_path):
    project = Project.from_path(project_path)
    env = Environment(project, load_plugins=False)
    output_path = tmp_path_factory.mktemp("webadmin-output")

    pad = Database(env).new_pad()
    builder = Builder(pad, output_path)
    builder.update_all_source_infos()

    app = WebAdmin(env, output_path=output_path)
    return app


@pytest.fixture
def test_client(webadmin):
    with webadmin.test_client() as client:
        yield client


def test_recordinfo_children_sort_limited_alts(test_client):
    # This exercises the bug described in #962, namely that
    # if a page has a child that only has content in a subset of the
    # configured alts, get_record_info throws an exception.
    data = test_client.get("/admin/api/recordinfo?path=/projects").get_json()
    child_data = data["children"]
    assert sorted(child_data, key=itemgetter("label")) == child_data


def test_eventstream_yield_bytes():
    count = 0

    @eventstream
    def testfunc():
        yield "string"
        yield 5

    for data in testfunc().response:  # pylint: disable=no-member
        count += 1
        assert isinstance(data, bytes)
    assert count >= 2


@pytest.mark.parametrize(
    "endpoint, params, invalid",
    [
        ("recordinfo", {}, {"path"}),
        ("previewinfo", {"path": "/", "alt": "BAD"}, {"alt"}),
    ],
)
def test_invalid_params(test_client, endpoint, params, invalid):
    resp = test_client.get(f"/admin/api/{endpoint}?{urlencode(params)}")
    assert resp.status_code == 400
    error = resp.get_json()["error"]
    assert error["title"] == "Invalid parameters"
    assert set(error["messages"].keys()) == invalid


def test_recordinfo(test_client):
    resp = test_client.get("/admin/api/recordinfo?path=%2F")
    assert resp.status_code == 200
    data = resp.get_json()
    assert any(att["id"] == "hello.txt" for att in data["attachments"])
    assert any(page["id"] == "blog" for page in data["children"])
    assert any(alt["alt"] == "de" for alt in data["alts"])


def test_delete_field(scratch_client, scratch_content_path):
    # None in page data means to delete the field
    # Test that that works
    contents_lr = scratch_content_path / "contents.lr"

    assert "\nbody:" in contents_lr.read_text()
    resp = scratch_client.put(
        "/admin/api/rawrecord?path=%2F", json={"path": "/", "data": {"body": None}}
    )
    assert resp.status_code == 200
    assert "\nbody:" not in contents_lr.read_text()


def test_get_path_info(test_client):
    resp = test_client.get("/admin/api/pathinfo?path=%2Fblog%2Fpost2")
    assert resp.get_json() == {
        "segments": [
            {
                "can_have_children": True,
                "exists": True,
                "id": "",
                "label_i18n": {"en": "Welcome"},
                "path": "/",
            },
            {
                "can_have_children": True,
                "exists": True,
                "id": "blog",
                "label_i18n": {"en": "Blog"},
                "path": "/blog",
            },
            {
                "can_have_children": True,
                "exists": True,
                "id": "post2",
                "label_i18n": {"en": "Post 2"},
                "path": "/blog/post2",
            },
        ],
    }


@pytest.mark.parametrize(
    "path, expect",
    [
        (
            "/blog/post1/hello.txt",
            {
                "exists": True,
                "url": "/blog/2015/12/post1/hello.txt",
                "is_hidden": False,
            },
        ),
        (
            "/extra/container",
            {
                "exists": True,
                "url": "/extra/container/",
                "is_hidden": True,
            },
        ),
        (
            "/missing",
            {
                "exists": False,
                "url": None,
                "is_hidden": True,
            },
        ),
    ],
)
def test_previewinfo(test_client, path, expect):
    resp = test_client.get(f"/admin/api/previewinfo?{urlencode({'path': path})}")
    assert resp.status_code == 200
    assert resp.get_json() == expect


@pytest.mark.parametrize("use_json", [True, False])
@pytest.mark.parametrize("lang", ["en", None])
def test_find(test_client, use_json, lang):
    # Test that we can pass params in JSON body, rather than in the query
    params = {"q": "hello", "alt": "_primary", "lang": lang}
    if use_json:
        resp = test_client.post("/admin/api/find", json=params)
    else:
        resp = test_client.post(f"/admin/api/find?{urlencode(params)}")
    assert resp.status_code == 200
    results = resp.get_json()["results"]
    assert any(result["title"] == "Hello" for result in results)
    assert len(results) == 1


@pytest.mark.parametrize(
    "path, alt, srcfile",
    [
        ("/projects/bagpipe", "de", "projects/bagpipe/contents+de.lr"),
        ("/hello.txt", "de", "hello.txt"),
    ],
)
def test_browsefs(test_client, mocker, project_path, path, alt, srcfile):
    launch = mocker.patch("click.launch")
    params = {"path": path, "alt": alt}
    resp = test_client.post("/admin/api/browsefs", json=params)
    assert resp.status_code == 200
    assert resp.get_json()["okay"]
    assert launch.mock_calls == [
        mocker.call(str(project_path / "content" / srcfile), locate=True),
    ]


@pytest.mark.parametrize(
    "path, alt, can_have_children",
    [
        ("/test.jpg", "de", False),
        ("/projects", "de", True),
    ],
)
def test_get_new_record_info(test_client, path, alt, can_have_children):
    params = {"path": path, "alt": alt}
    resp = test_client.get(f"/admin/api/newrecord?{urlencode(params)}")
    assert resp.status_code == 200
    data = resp.get_json()
    assert data["can_have_children"] is bool(can_have_children)


@pytest.mark.parametrize(
    "path, alt, can_upload",
    [
        ("/test.jpg", "de", False),
        ("/projects", "de", True),
    ],
)
def test_get_new_attachment_info(test_client, path, alt, can_upload):
    params = {"path": path}
    resp = test_client.get(f"/admin/api/newattachment?{urlencode(params)}")
    assert resp.status_code == 200
    data = resp.get_json()
    assert data["can_upload"] is bool(can_upload)


def test_upload_new_attachment(scratch_client, scratch_content_path):
    params = {
        "path": "/page",
        "file": (BytesIO(b"foo data"), "foo.txt"),
    }
    resp = scratch_client.post("/admin/api/newattachment", data=params)
    assert resp.status_code == 200
    assert not resp.get_json()["bad_upload"]
    dstpath = scratch_content_path / "page/foo.txt"
    assert dstpath.read_bytes() == b"foo data"


@pytest.mark.parametrize(
    "path, alt",
    [
        ("/test.txt", PRIMARY_ALT),
        ("/missing", PRIMARY_ALT),
        ("/missing", "de"),
    ],
)
def test_upload_new_attachment_failure(scratch_client, scratch_content_path, path, alt):
    scratch_content_path.joinpath("test.txt").write_bytes(b"test")
    params = {
        "path": path,
        "alt": alt,
        "file": (BytesIO(b"foo data"), "foo.txt"),
    }
    resp = scratch_client.post("/admin/api/newattachment", data=params)
    assert resp.status_code == 200
    assert resp.get_json()["bad_upload"]
    dstpath = scratch_content_path / "test.txt/foo.txt"
    assert not dstpath.exists()


@pytest.mark.parametrize(
    "path, id, expect, creates",
    [
        (
            "/page",
            "new",
            {"valid_id": True, "exists": False, "path": "/page/new"},
            "page/new/contents.lr",
        ),
        ("/page", ".new", {"valid_id": False, "exists": False, "path": None}, None),
        ("/", "page", {"valid_id": True, "exists": True, "path": "/page"}, None),
    ],
)
def test_add_new_record(scratch_client, scratch_project, path, id, expect, creates):
    def tree_files():
        return {
            Path(dirpath, filename)
            for (dirpath, dirnames, filenames) in os.walk(scratch_project.tree)
            for filename in filenames
        }

    orig_tree_files = tree_files()
    params = {"path": path, "id": id, "data": {}}
    resp = scratch_client.post("/admin/api/newrecord", json=params)
    assert resp.status_code == 200
    assert resp.get_json() == expect
    new_files = tree_files() - orig_tree_files
    if creates is None:
        assert len(new_files) == 0
    else:
        assert new_files == {Path(scratch_project.tree, "content", creates)}


@pytest.mark.parametrize(
    "path, id",
    [
        # Ensure that attempts to create records outside of the `content` subtree fail.
        # (Reported by Riku Bamba)
        ("/../../templates", ""),
    ],
)
def test_add_new_record_bad_request(scratch_client, scratch_project, path, id):
    params = {"path": path, "id": id, "data": {}}
    resp = scratch_client.post("/admin/api/newrecord", json=params)
    assert resp.status_code == 400
    assert resp.get_json() == {
        "error": {
            "title": "Invalid parameters",
            "messages": {"path": ["Invalid value."]},
        },
    }


@pytest.mark.parametrize(
    "delete_master, expect", [("1", True), (True, True), ("0", False), (False, False)]
)
def test_delete_record(scratch_client, mocker, delete_master, expect):
    delete = mocker.patch.object(EditorSession, "delete")
    params = {"path": "/myobj", "delete_master": delete_master}
    resp = scratch_client.post("/admin/api/deleterecord", json=params)
    assert resp.status_code == 200
    assert delete.mock_calls == [
        mocker.call(delete_master=expect),
    ]


@pytest.mark.parametrize(
    "url_path, expect",
    [
        (
            "/blog/2015/12/post1/hello.txt",
            {
                "exists": True,
                "path": "/blog/post1/hello.txt",
                "alt": PRIMARY_ALT,
            },
        ),
        (
            "/missing",
            {
                "exists": False,
                "path": None,
                "alt": None,
            },
        ),
        (
            "/static/demo.css",  # non-Records do not exist
            {
                "exists": False,
                "path": None,
                "alt": None,
            },
        ),
    ],
)
def test_match_url(test_client, url_path, expect):
    params = {"url_path": url_path}
    resp = test_client.get(f"/admin/api/matchurl?{urlencode(params)}")
    assert resp.status_code == 200
    assert resp.get_json() == expect


def test_get_raw_records(test_client):
    resp = test_client.get("/admin/api/rawrecord?path=%2F")
    assert resp.status_code == 200
    data = resp.get_json()
    assert data["data"]["title"] == "Welcome"
    assert "datamodel" in data


def test_servers(test_client):
    resp = test_client.get("/admin/api/servers")
    assert resp.status_code == 200
    assert any(server["id"] == "production" for server in resp.get_json()["servers"])


def test_build(test_client, webadmin, mocker):
    builder = mocker.patch.object(LektorContext, "builder")
    resp = test_client.post("/admin/api/build")
    assert resp.status_code == 200
    data = resp.get_json()
    assert data["okay"]
    assert builder.mock_calls == [
        mocker.call.build_all(),
        mocker.call.prune(),
    ]


def test_clean(test_client, webadmin, mocker):
    builder = mocker.patch.object(LektorContext, "builder")
    resp = test_client.post("/admin/api/clean")
    assert resp.status_code == 200
    data = resp.get_json()
    assert data["okay"]
    assert mocker.call.prune(all=True) in builder.mock_calls


@pytest.mark.xfail(reason="FIXME")
def test_publish_requires_post(test_client, mocker):
    # See https://github.com/lektor/lektor/issues/1006
    resp = test_client.get("/admin/api/publish?server=unknown")
    assert resp.status_code == 405


def test_publish(test_client, mocker):
    def dummy_publish(env, target, output_path, credentials=None, **extra):
        yield "line1"
        raise PublishError("wups")

    mocker.patch("lektor.admin.modules.api.publish", side_effect=dummy_publish)

    resp = test_client.get("/admin/api/publish?server=production")
    assert resp.status_code == 200
    assert list(resp.response) == [
        b'data: {"msg": "line1"}\n\n',
        b'data: {"msg": "Error: wups"}\n\n',
        b"data: null\n\n",
    ]


@pytest.mark.parametrize(
    "params",
    [
        {"server": "bogus"},
        {},
    ],
)
def test_publish_bad_params(test_client, params):
    resp = test_client.get(f"/admin/api/publish?{urlencode(params)}")
    assert resp.status_code == 400
    assert "server" in resp.get_json()["error"]["messages"]


def test_ping(test_client):
    resp = test_client.get("/admin/api/ping")
    assert resp.status_code == 200
    data = resp.get_json()
    assert data["okay"]


@pytest.mark.parametrize(
    "url",
    [
        "/missing.txt",
        "/admin/missing.txt",
        "/admin/api/missing.txt",
        "/admin/api/edit/missing.txt",
    ],
)
def test_missing_files(test_client, url):
    resp = test_client.get(url)
    assert resp.status_code == 404


@pytest.mark.parametrize(
    "url, allowed",
    [
        ("/", {"GET"}),
        ("/admin/api/recordinfo?path=%2F", {"GET"}),
        ("/admin/api/clean", {"POST"}),
        ("/admin/api/rawrecord", {"GET", "PUT"}),
    ],
)
@pytest.mark.parametrize("method", ["GET", "POST", "PUT"])
def test_allowed_methods(test_client, method, url, allowed):
    if method not in allowed:
        extra_allowed = {"OPTIONS", "HEAD"} if "GET" in allowed else {"OPTIONS"}
        resp = test_client.open(url, method=method)
        assert resp.status_code == 405
        assert set(resp.allow) == allowed | extra_allowed



from functools import partial

import pytest

from lektor.build_programs import FileAssetBuildProgram
from lektor.builder import Builder
from lektor.reporter import describe_build_func


@pytest.fixture
def build_state(pad, tmp_path):
    builder = Builder(pad, destination_path=tmp_path)
    return builder.new_build_state()


def dummy_build_func(*args):
    pass


def test_describe_build_func():
    assert describe_build_func(dummy_build_func) == "test_reporter.dummy_build_func"


def test_describe_build_func_describes_partial():
    build_func = partial(dummy_build_func, "dummy-arg")
    assert describe_build_func(build_func) == "test_reporter.dummy_build_func"


def test_describe_build_func_BuildProgram(pad, build_state):
    build_program = FileAssetBuildProgram(pad.get_asset("static/demo.css"), build_state)
    build_func = build_program.build_artifact
    assert (
        describe_build_func(build_func) == "lektor.build_programs.FileAssetBuildProgram"
    )


def test_describe_build_func_deals_with_garbage():
    garbage = object()
    assert isinstance(describe_build_func(garbage), str)



from __future__ import annotations

import errno
import hashlib
import io
import os
import posixpath
import urllib.parse
from collections.abc import Callable
from collections.abc import Generator
from collections.abc import Iterable
from collections.abc import Iterator
from collections.abc import Mapping
from collections.abc import Sequence
from contextlib import AbstractContextManager
from contextlib import contextmanager
from contextlib import ExitStack
from contextlib import suppress
from ftplib import Error as FTPError
from inspect import cleandoc
from pathlib import Path
from subprocess import CalledProcessError
from subprocess import CompletedProcess
from subprocess import DEVNULL
from subprocess import PIPE
from subprocess import STDOUT
from tempfile import TemporaryDirectory
from typing import Any
from typing import NoReturn
from typing import TYPE_CHECKING
from urllib.parse import urlsplit
from warnings import warn

from werkzeug.datastructures import MultiDict

from lektor.compat import werkzeug_urls_URL
from lektor.exception import LektorException
from lektor.utils import bool_from_string
from lektor.utils import locate_executable
from lektor.utils import portable_popen


if TYPE_CHECKING:  # pragma: no cover
    from _typeshed import StrOrBytesPath
    from _typeshed import StrPath

    from lektor.environment import Environment


def _parse_query(query: str, **kwargs: Any) -> MultiDict:
    return MultiDict(urllib.parse.parse_qsl(query, **kwargs))


def _ascii_host(host: str) -> str:
    """Translate internationalized domain name to IDNA-encoded ASCII."""
    return host.encode("idna").decode("ascii")


@contextmanager
def _ssh_key_file(credentials: Mapping[str, str] | None) -> Iterator[StrPath | None]:
    with ExitStack() as stack:
        key_file: StrPath | None
        key_file = credentials.get("key_file") if credentials else None
        key = credentials.get("key") if credentials else None
        if not key_file and key:
            if ":" in key:
                key_type, _, key = key.partition(":")
                key_type = key_type.upper()
            else:
                key_type = "RSA"
            key_file = Path(stack.enter_context(TemporaryDirectory()), "keyfile")
            with key_file.open("w", encoding="utf-8") as f:
                f.write(f"-----BEGIN {key_type} PRIVATE KEY-----\n")
                f.writelines(key[x : x + 64] + "\n" for x in range(0, len(key), 64))
                f.write(f"-----END {key_type} PRIVATE KEY-----\n")
        yield key_file


@contextmanager
def _ssh_command(
    credentials: Mapping[str, str] | None, port: int | None = None
) -> Iterator[str | None]:
    with _ssh_key_file(credentials) as key_file:
        args = []
        if port:
            args.append(f" -p {port}")
        if key_file:
            args.append(f' -i "{key_file}" -o IdentitiesOnly=yes')
        if args:
            ssh_command = "ssh" + " ".join(args)
        else:
            ssh_command = None

        yield ssh_command


class PublishError(LektorException):
    """Raised by publishers if something goes wrong."""


class Command(AbstractContextManager["Command"]):
    """A wrapper around subprocess.Popen to facilitate streaming output via generator.

    :param argline: Command with arguments to execute.
    :param cwd: Optional. Directory in which to execute command.
    :param env: Optional. Environment with which to run command.
    :param capture: Default `True`. Whether to capture stdout and stderr.
    :param silent: Default `False`. Discard output altogether.
    :param check: Default `False`.
        If set, raise ``CalledProcessError`` on non-zero return code.
    :param input: Optional. A string to feed to the subprocess via stdin.
    :param capture_stdout: Default `False`. Capture stdout and
        return in ``CompletedProcess.stdout``.

    Basic Usage
    ===========

    To run a command, returning any output on stdout or stderr to the caller
    as an iterable (generator), while checking the return code from the command:

        def run_command(argline):
            # This passes the output
            rv = yield from Command(argline)
            if rv.returncode != 0:
                raise RuntimeError("Command failed!")

    This could be called as follows:

        for outline in run_command(('ls')):
            print(outline.rstrip())

    Supplying input via stdin, Capturing stdout
    ===========================================

    The following example shows how input may be fed to a subprocess via stdin,
    and how stdout may be captured for further processing.

        def run_wc(input):
            rv = yield from Command(
                ('wc'), check=True, input=input, capture_stdout=True
            )
            lines, words, chars = rv.stdout.split()
            print(f"{words} words, {chars} chars")

        stderr_lines = list(run_wc("a few words"))
        # prints "3 words, 11 chars"

    Note that ``check=True`` will cause a ``CalledProcessError`` to be raised if the
    ``wc`` subprocess returns a non-zero return code.

    """

    def __init__(
        self,
        argline: Iterable[str],
        *,
        cwd: StrOrBytesPath | None = None,
        env: Mapping[str, str] | None = None,
        capture: bool = True,
        silent: bool = False,
        check: bool = False,
        input: str | None = None,
        capture_stdout: bool = False,
    ) -> None:
        kwargs: dict[str, Any] = {"cwd": cwd}
        if env:
            kwargs["env"] = {**os.environ, **env}
        if silent:
            kwargs["stdout"] = DEVNULL
            kwargs["stderr"] = DEVNULL
            capture = False
        if input is not None:
            kwargs["stdin"] = PIPE
        if capture or capture_stdout:
            kwargs["stdout"] = PIPE
        if capture:
            kwargs["stderr"] = STDOUT if not capture_stdout else PIPE

        # Python >= 3.7 has sane encoding defaults in the case that the system is
        # (likely mis-)configured to use ASCII as the default encoding (PEP538).
        # It also provides a way for the user to force the use of UTF-8 (PEP540).
        kwargs["text"] = True
        kwargs["errors"] = "replace"

        self.capture = capture  # b/c - unused
        self.check = check
        self._stdout = None

        with ExitStack() as stack:
            self._cmd = stack.enter_context(portable_popen(list(argline), **kwargs))
            self._closer: Callable[[], None] | None = stack.pop_all().close

        if input is not None or capture_stdout:
            self._output = self._communicate(input, capture_stdout, capture)
        elif capture:
            self._output = self._cmd.stdout
        else:
            self._output = None

    def _communicate(
        self, input: str | None, capture_stdout: bool, capture: bool
    ) -> Iterator[str] | None:
        proc = self._cmd
        try:
            if capture_stdout:
                self._stdout, errout = proc.communicate(input)
            else:
                errout, _ = proc.communicate(input)
        except BaseException:
            proc.kill()
            with suppress(CalledProcessError):
                self.close()
            raise
        if capture:
            return iter(errout.splitlines())
        return None

    def close(self) -> None:
        """Wait for subprocess to complete.

        If check=True was passed to the constructor, raises ``CalledProcessError``
        if the subprocess returns a non-zero status code.
        """
        closer, self._closer = self._closer, None
        if closer:
            # This waits for process and closes standard file descriptors
            closer()
        if self.check:
            rc = self._cmd.poll()
            if rc != 0:
                raise CalledProcessError(rc, self._cmd.args, self._stdout)

    def wait(self) -> int:
        """Wait for subprocess to complete. Return status code."""
        self._cmd.wait()
        self.close()
        return self._cmd.returncode

    def result(self) -> CompletedProcess[str]:
        """Wait for subprocess to complete.  Return ``CompletedProcess`` instance.

        If ``capture_stdout=True`` was passed to the constructor, the output
        captured from stdout will be available on the ``.stdout`` attribute
        of the return value.
        """
        return CompletedProcess(self._cmd.args, self.wait(), self._stdout)

    @property
    def returncode(self) -> int | None:
        """Return exit status of the subprocess.

        Or ``None`` if the subprocess is still alive.
        """
        return self._cmd.returncode

    def __exit__(self, *__: Any) -> None:
        self.close()

    def __iter__(self) -> Generator[str, None, CompletedProcess[str]]:
        """A generator with yields any captured output and returns a
        ``CompletedProcess``.

        If ``capture`` is ``True`` (the default).  Both stdout and stderr are available
        in the iterator output.

        If ``capture_stdout`` is set, stdout is captured to a string which is made
        available via ``CompletedProcess.stdout`` attribute of the return value. Stderr
        output is available via the iterator output, as normal.
        """
        if self._output is None:
            raise RuntimeError("Not capturing")
        for line in self._output:
            yield line.rstrip()
        return self.result()

    safe_iter = __iter__  # b/c - deprecated

    @property
    def output(self) -> Iterator[str]:  # b/c - deprecated
        return self.safe_iter()


class Publisher:
    def __init__(self, env: Environment, output_path: str) -> None:
        self.env = env
        self.output_path = os.path.abspath(output_path)

    def fail(self, message: str) -> NoReturn:
        # pylint: disable=no-self-use
        raise PublishError(message)

    def publish(
        self,
        target_url: str,
        credentials: Mapping[str, str] | None = None,
        **extra: Any,
    ) -> Iterator[str]:
        raise NotImplementedError()


class RsyncPublisher(Publisher):
    @contextmanager
    def get_command(self, target_url, credentials):
        credentials = credentials or {}
        argline = ["rsync", "-rclzv", "--exclude=.lektor"]
        target = []
        env = {}

        url = urlsplit(target_url)
        options = _parse_query(url.query, keep_blank_values=True)
        exclude = options.getlist("exclude")
        for file in exclude:
            argline.extend(("--exclude", file))

        delete = options.get("delete", False) in ("", "on", "yes", "true", "1", None)
        if delete:
            argline.append("--delete-after")

        with _ssh_command(credentials, url.port) as ssh_command:
            if ssh_command:
                argline.extend(("-e", ssh_command))

            username = credentials.get("username") or url.username
            if username:
                target.append(username + "@")

            if url.hostname is not None:
                target.append(_ascii_host(url.hostname))
                target.append(":")
            target.append(url.path.rstrip("/") + "/")

            argline.append(self.output_path.rstrip("/\\") + "/")
            argline.append("".join(target))
            yield Command(argline, env=env)

    def publish(self, target_url, credentials=None, **extra):
        with self.get_command(target_url, credentials) as client:
            yield from client


class FtpConnection:
    def __init__(self, target_url, credentials=None):
        credentials = credentials or {}
        url = urlsplit(target_url)
        if url.hostname is None:
            raise PublishError(
                "No host name was specified in the target URL ({target_url})"
            )
        self.con = self.make_connection()
        self.url = url
        self.username = credentials.get("username") or url.username
        self.password = credentials.get("password") or url.password
        self.log_buffer = []
        self._known_folders = set()

    @staticmethod
    def make_connection():
        # pylint: disable=import-outside-toplevel
        from ftplib import FTP

        return FTP()

    def drain_log(self):
        log = self.log_buffer[:]
        del self.log_buffer[:]
        for chunk in log:
            for line in chunk.splitlines():
                if not isinstance(line, str):
                    line = line.decode("utf-8", "replace")
                yield line.rstrip()

    def connect(self):
        options = _parse_query(self.url.query, keep_blank_values=True)
        assert self.url.hostname is not None
        host = _ascii_host(self.url.hostname)
        port = self.url.port or 21

        log = self.log_buffer
        log.append("000 Connecting to server ...")
        try:
            log.append(self.con.connect(host, port))
        except Exception as e:
            log.append("000 Could not connect.")
            log.append(str(e))
            return False

        try:
            credentials = {}
            if self.username:
                credentials["user"] = self.username
            if self.password:
                credentials["passwd"] = self.password
            log.append(self.con.login(**credentials))

        except Exception as e:
            log.append("000 Could not authenticate.")
            log.append(str(e))
            return False

        passive = options.get("passive") in ("on", "yes", "true", "1", None)
        log.append("000 Using passive mode: %s" % (passive and "yes" or "no"))
        self.con.set_pasv(passive)

        try:
            log.append(self.con.cwd(self.url.path))
        except Exception as e:
            log.append(str(e))
            return False

        log.append("000 Connected!")
        return True

    def mkdir(self, path, recursive=True):
        if not isinstance(path, str):
            path = path.decode("utf-8")
        if path in self._known_folders:
            return
        dirname, _ = posixpath.split(path)
        if dirname and recursive:
            self.mkdir(dirname)
        try:
            self.con.mkd(path)
        except FTPError as e:
            msg = str(e)
            if msg[:4] != "550 ":
                self.log_buffer.append(str(e))
                return
        self._known_folders.add(path)

    def append(self, filename, data):
        if not isinstance(filename, str):
            filename = filename.decode("utf-8")

        input = io.BytesIO(data.encode("utf-8"))

        try:
            self.con.storbinary("APPE " + filename, input)
        except FTPError as e:
            self.log_buffer.append(str(e))
            return False
        return True

    def get_file(self, filename, out=None):
        if not isinstance(filename, str):
            filename = filename.decode("utf-8")
        getvalue = False
        if out is None:
            out = io.BytesIO()
            getvalue = True
        try:
            self.con.retrbinary("RETR " + filename, out.write)
        except FTPError as e:
            msg = str(e)
            if msg[:4] != "550 ":
                self.log_buffer.append(e)
            return None
        if getvalue:
            return out.getvalue().decode("utf-8")
        return out

    def upload_file(self, filename, src, mkdir=False):
        if isinstance(src, str):
            src = io.BytesIO(src.encode("utf-8"))
        if mkdir:
            directory = posixpath.dirname(filename)
            if directory:
                self.mkdir(directory, recursive=True)
        if not isinstance(filename, str):
            filename = filename.decode("utf-8")
        try:
            self.con.storbinary("STOR " + filename, src, blocksize=32768)
        except FTPError as e:
            self.log_buffer.append(str(e))
            return False
        return True

    def rename_file(self, src, dst):
        try:
            self.con.rename(src, dst)
        except FTPError as e:
            self.log_buffer.append(str(e))
            try:
                self.con.delete(dst)
            except Exception as e:
                self.log_buffer.append(str(e))
            try:
                self.con.rename(src, dst)
            except Exception as e:
                self.log_buffer.append(str(e))

    def delete_file(self, filename):
        if isinstance(filename, str):
            filename = filename.encode("utf-8")
        try:
            self.con.delete(filename)
        except Exception as e:
            self.log_buffer.append(str(e))

    def delete_folder(self, filename):
        if isinstance(filename, str):
            filename = filename.encode("utf-8")
        try:
            self.con.rmd(filename)
        except Exception as e:
            self.log_buffer.append(str(e))
        self._known_folders.discard(filename)


class FtpTlsConnection(FtpConnection):
    @staticmethod
    def make_connection():
        # pylint: disable=import-outside-toplevel
        from ftplib import FTP_TLS

        return FTP_TLS()

    def connect(self):
        connected = super().connect()
        if connected:
            # Upgrade data connection to TLS.
            self.con.prot_p()  # pylint: disable=no-member
        return connected


class FtpPublisher(Publisher):
    connection_class = FtpConnection

    @staticmethod
    def read_existing_artifacts(con):
        contents = con.get_file(".lektor/listing")
        if not contents:
            return {}, set()
        duplicates = set()
        rv = {}
        # Later records override earlier ones.  There can be duplicate
        # entries if the file was not compressed.
        for line in contents.splitlines():
            items = line.split("|")
            if len(items) == 2:
                if not isinstance(items[0], str):
                    artifact_name = items[0].decode("utf-8")
                else:
                    artifact_name = items[0]
                if artifact_name in rv:
                    duplicates.add(artifact_name)
                rv[artifact_name] = items[1]
        return rv, duplicates

    def iter_artifacts(self):
        """Iterates over all artifacts in the build folder and yields the
        artifacts.
        """
        for dirpath, dirnames, filenames in os.walk(self.output_path):
            dirnames[:] = [x for x in dirnames if not self.env.is_ignored_artifact(x)]
            for filename in filenames:
                if self.env.is_ignored_artifact(filename):
                    continue
                full_path = os.path.join(self.output_path, dirpath, filename)
                local_path = full_path[len(self.output_path) :].lstrip(os.path.sep)
                if os.path.altsep:
                    local_path = local_path.lstrip(os.path.altsep)
                h = hashlib.sha1()
                try:
                    with open(full_path, "rb") as f:
                        while 1:
                            item = f.read(4096)
                            if not item:
                                break
                            h.update(item)
                except OSError as e:
                    if e.errno != errno.ENOENT:
                        raise
                yield (
                    local_path.replace(os.path.sep, "/"),
                    full_path,
                    h.hexdigest(),
                )

    @staticmethod
    def get_temp_filename(filename):
        dirname, basename = posixpath.split(filename)
        return posixpath.join(dirname, "." + basename + ".tmp")

    def upload_artifact(self, con, artifact_name, source_file, checksum):
        with open(source_file, "rb") as source:
            tmp_dst = self.get_temp_filename(artifact_name)
            con.log_buffer.append(f"000 Updating {artifact_name}")
            con.upload_file(tmp_dst, source, mkdir=True)
            con.rename_file(tmp_dst, artifact_name)
            con.append(".lektor/listing", f"{artifact_name}|{checksum}\n")

    def consolidate_listing(self, con, current_artifacts):
        server_artifacts, duplicates = self.read_existing_artifacts(con)
        known_folders = set()
        for artifact_name in current_artifacts.keys():
            known_folders.add(posixpath.dirname(artifact_name))

        for artifact_name in server_artifacts:
            if artifact_name not in current_artifacts:
                con.log_buffer.append(f"000 Deleting {artifact_name}")
                con.delete_file(artifact_name)
                folder = posixpath.dirname(artifact_name)
                if folder not in known_folders:
                    con.log_buffer.append(f"000 Deleting {folder}")
                    con.delete_folder(folder)

        if duplicates or server_artifacts != current_artifacts:
            listing = []
            for artifact_name, checksum in current_artifacts.items():
                listing.append(f"{artifact_name}|{checksum}\n")
            listing.sort()
            con.upload_file(".lektor/.listing.tmp", "".join(listing))
            con.rename_file(".lektor/.listing.tmp", ".lektor/listing")

    def publish(self, target_url, credentials=None, **extra):
        con = self.connection_class(target_url, credentials)
        connected = con.connect()
        yield from con.drain_log()
        if not connected:
            return

        yield "000 Reading server state ..."
        con.mkdir(".lektor")
        committed_artifacts, _ = self.read_existing_artifacts(con)
        yield from con.drain_log()

        yield "000 Begin sync ..."
        current_artifacts = {}
        for artifact_name, filename, checksum in self.iter_artifacts():
            current_artifacts[artifact_name] = checksum
            if checksum != committed_artifacts.get(artifact_name):
                self.upload_artifact(con, artifact_name, filename, checksum)
                yield from con.drain_log()
        yield "000 Sync done!"

        yield "000 Consolidating server state ..."
        self.consolidate_listing(con, current_artifacts)
        yield from con.drain_log()

        yield "000 All done!"


class FtpTlsPublisher(FtpPublisher):
    connection_class = FtpTlsConnection


class GitRepo(AbstractContextManager["GitRepo"]):
    """A temporary git repository.

    This class provides some lower-level utility methods which may be
    externally useful, but the main use case is:

        def publish(html_output):
            gitrepo = GitRepo(html_output)
            yield from gitrepo.publish_ghpages(
                push_url="git@github.com:owner/repo.git",
                branch="gh-pages"
            )

    :param work_tree: The work tree for the repository.
    """

    def __init__(self, work_tree: StrPath) -> None:
        environ = {**os.environ, "GIT_WORK_TREE": str(work_tree)}

        for what, default in [("NAME", "Lektor Bot"), ("EMAIL", "bot@getlektor.com")]:
            value = (
                environ.get(f"GIT_AUTHOR_{what}")
                or environ.get(f"GIT_COMMITTER_{what}")
                or default
            )
            for key in f"GIT_AUTHOR_{what}", f"GIT_COMMITTER_{what}":
                environ[key] = environ.get(key) or value

        with ExitStack() as stack:
            environ["GIT_DIR"] = stack.enter_context(TemporaryDirectory(suffix=".git"))
            self.environ = environ
            self.run("init", "--quiet")

            self._exit_stack = stack.pop_all()

    def __exit__(self, *__: Any) -> None:
        self._exit_stack.close()

    def _popen(self, args: Sequence[str], **kwargs: Any) -> Command:
        cmd = ["git"]
        cmd.extend(args)
        return Command(cmd, env=self.environ, **kwargs)

    def popen(
        self,
        *args: str,
        check: bool = True,
        input: str | None = None,
        capture_stdout: bool = False,
    ) -> Command:
        """Run a git subcommand."""
        return self._popen(
            args, check=check, input=input, capture_stdout=capture_stdout
        )

    def run(
        self,
        *args: str,
        check: bool = True,
        input: str | None = None,
        capture_stdout: bool = False,
    ) -> CompletedProcess[str]:
        """Run a git subcommand and wait for completion."""
        return self._popen(
            args, check=check, input=input, capture_stdout=capture_stdout, capture=False
        ).result()

    def set_ssh_credentials(self, credentials: Mapping[str, str]) -> None:
        """Set up git ssh credentials.

        This repository will be configured to used whatever SSH credentials
        can found in ``credentials`` (if any).
        """
        stack = self._exit_stack
        ssh_command = stack.enter_context(_ssh_command(credentials))
        if ssh_command:
            self.environ.setdefault("GIT_SSH_COMMAND", ssh_command)

    def set_https_credentials(self, credentials: Mapping[str, str]) -> None:
        """Set up git http(s) credentials.

        This repository will be configured to used whatever HTTP credentials
        can found in ``credentials`` (if any).
        """
        username = credentials.get("username", "")
        password = credentials.get("password")
        if username or password:
            userpass = f"{username}:{password}" if password else username
            git_dir = self.environ["GIT_DIR"]
            cred_file = Path(git_dir, "lektor_cred_file")
            # pylint: disable=unspecified-encoding
            cred_file.write_text(f"https://{userpass}@github.com\n")
            self.run("config", "credential.helper", f'store --file "{cred_file}"')

    def add_to_index(self, filename: str, content: str) -> None:
        """Create a file in the index.

        This creates file named ``filename`` with content ``content`` in the git
        index.
        """
        oid = self.run(
            "hash-object", "-w", "--stdin", input=content, capture_stdout=True
        ).stdout.strip()
        self.run("update-index", "--add", "--cacheinfo", "100644", oid, filename)

    def publish_ghpages(
        self,
        push_url: str,
        branch: str,
        cname: str | None = None,
        preserve_history: bool = True,
    ) -> Iterator[str]:
        """Publish the contents of the work tree to GitHub pages.

        :param push_url: The URL to push to.
        :param branch: The branch to push to
        :param cname: Optional. Create a top-level ``CNAME`` with given contents.
        """
        refspec = f"refs/heads/{branch}"
        if preserve_history:
            yield "Fetching existing head"
            fetch_cmd = self.popen("fetch", "--depth=1", push_url, refspec, check=False)
            yield from _prefix_output(fetch_cmd)
            if fetch_cmd.returncode == 0:
                # If fetch was succesful, reset HEAD to remote head
                yield from _prefix_output(self.popen("reset", "--soft", "FETCH_HEAD"))
            else:
                # otherwise assume remote branch does not exist
                yield f"Creating new branch {branch}"

        # At this point, the index is still empty. Add all but .lektor dir to index
        yield from _prefix_output(
            self.popen("add", "--force", "--all", "--", ".", ":(exclude).lektor")
        )
        if cname is not None:
            self.add_to_index("CNAME", f"{cname}\n")

        # Check for changes
        diff_cmd = self.popen(
            "diff", "--cached", "--no-renames", "--exit-code", "--quiet", check=False
        )
        yield from _prefix_output(diff_cmd)
        if diff_cmd.returncode == 0:
            yield "No changes to publish☺"
        elif diff_cmd.returncode == 1:
            yield "Creating commit"
            yield from _prefix_output(
                self.popen("commit", "--quiet", "--message", "Synchronized build")
            )
            push_cmd = ["push", push_url, f"HEAD:{refspec}"]
            if not preserve_history:
                push_cmd.insert(1, "--force")
            yield "Pushing to github"
            yield from _prefix_output(self.popen(*push_cmd))
            yield "Success!"
        else:
            diff_cmd.result().check_returncode()  # raise error


def _prefix_output(lines: Iterable[str], prefix: str = "> ") -> Iterator[str]:
    """Add prefix to lines."""
    return (f"{prefix}{line}" for line in lines)


class GithubPagesPublisher(Publisher):
    """Publish to GitHub pages."""

    def publish(
        self,
        target_url: str,
        credentials: Mapping[str, str] | None = None,
        **extra: Any,
    ) -> Iterator[str]:
        if not locate_executable("git"):
            self.fail("git executable not found; cannot deploy.")

        push_url, branch, cname, preserve_history, warnings = self._parse_url(
            target_url
        )
        creds = self._parse_credentials(credentials, target_url)

        yield from iter(warnings)

        with GitRepo(self.output_path) as repo:
            if push_url.startswith("https:"):
                repo.set_https_credentials(creds)
            else:
                repo.set_ssh_credentials(creds)
            yield from repo.publish_ghpages(push_url, branch, cname, preserve_history)

    def _parse_url(
        self, target_url: str
    ) -> tuple[str, str, str | None, bool, Sequence[str]]:
        url = urlsplit(target_url)
        if not url.hostname:
            self.fail("github owner missing from target URL")
        gh_owner = url.hostname.lower()
        gh_project = url.path.strip("/").lower()
        if not gh_project:
            self.fail("github project missing from target URL")

        params = _parse_query(url.query, keep_blank_values=True)
        cname = params.get("cname")
        branch = params.get("branch")
        preserve_history = bool_from_string(params.get("preserve_history"), True)

        warnings = []

        if not branch:
            if gh_project == f"{gh_owner}.github.io":
                warnings.extend(
                    cleandoc(self._EXPLICIT_BRANCH_SUGGESTED_MSG).splitlines()
                )
                warn(
                    " ".join(
                        cleandoc(self._DEFAULT_BRANCH_DEPRECATION_MSG).splitlines()
                    ),
                    # deprecated in version 3.4.0
                    category=FutureWarning,
                    stacklevel=1,
                )
                branch = "master"
            else:
                branch = "gh-pages"

        if url.scheme in ("ghpages", "ghpages+ssh"):
            push_url = f"ssh://git@github.com/{gh_owner}/{gh_project}.git"
            default_port = 22
        else:
            push_url = f"https://github.com/{gh_owner}/{gh_project}.git"
            default_port = 443
        if url.port and url.port != default_port:
            self.fail("github does not support pushing to non-standard ports")

        return push_url, branch, cname, preserve_history, warnings

    _EXPLICIT_BRANCH_SUGGESTED_MSG = """
    ================================================================
    WARNING!!! You should explicitly set the name of the published
    branch of your GitHub pages repository.

    The default branch for new GitHub pages repositories has changed
    to 'main', but Lektor still defaults to the old value, 'master'.
    In a future version of Lektor, the default branch name will
    changed to match the new GitHub default.

    For details, see
    https://getlektor.com/docs/deployment/ghpages/#pushing-to-an-explicit-branch
    ================================================================
    """

    _DEFAULT_BRANCH_DEPRECATION_MSG = """
    Currently, by default, Lektor pushes to the 'master' branch when
    deploying to GitHub pages repositories.  In a future version of
    Lektor, the default branch will GitHub's new default, 'main'.
    It is suggested that you explicitly set which branch to push to.
    """

    @staticmethod
    def _parse_credentials(
        credentials: Mapping[str, str] | None, target_url: str
    ) -> Mapping[str, str]:
        url = urlsplit(target_url)
        creds = dict(credentials or {})
        # Fill in default username/password from target url
        for key, default in [
            ("username", url.username),
            ("password", url.password),
        ]:
            if not creds.get(key) and default:
                creds[key] = default
        return creds


builtin_publishers = {
    "rsync": RsyncPublisher,
    "ftp": FtpPublisher,
    "ftps": FtpTlsPublisher,
    "ghpages": GithubPagesPublisher,
    "ghpages+https": GithubPagesPublisher,
    "ghpages+ssh": GithubPagesPublisher,
}


def publish(env, target, output_path, credentials=None, **extra):
    target_url = _CompatURLStr(target)
    url = urlsplit(target_url)
    publisher = env.publishers.get(url.scheme)
    if publisher is None:
        raise PublishError(f'"{url.scheme}" is an unknown scheme.')
    return publisher(env, output_path).publish(target_url, credentials, **extra)


class _CompatURLStr(str):
    """A string that provides some features of the werkzeug.urls.URL split URL class.

    We used to pass a ``werkzeug.urls.URL`` instance as the ``target_url`` argument to
    the ``Publisher.publish`` method.  Werkzeug has deprecated the ``URL`` class, so
    now we've changed our API to just pass a ``str`` for ``target_url``.

    There are however, Lektor plugins out in the wild that provide their own custom
    Publisher classes, and they expect a ``URL`` instance for ``target_url``.

    Here we provide most of the methods and attributes of ``URL`` that might be of use
    to a publisher, so as to try not to break all those existing plugins.

    .. tip::
       New plugins may preserve compatibility with older versions of Lektor by
       first coercing their ``target_url`` parameter to a ``str`` before use.
       (This works because ``werkzeug.urls.URL.__str__`` returns the reassembled URL.)
       E.g. using ``urllib.parse.urlsplit`` to parse the URL:

       .. code:: python
         from urllib.parse import urlsplit

         class CustomPublisher(Publisher):
             def publish(self, target_url, credentials=None, **extra):
                 url = urlsplit(str(target_url))
                 host = url.hostname
                 ...
    """

    def __getattr__(self, name: str):
        if name.startswith("_"):
            raise AttributeError(name)
        url = werkzeug_urls_URL(*urlsplit(self))
        rv = getattr(url, name)
        warn(
            "Since Lektor version 3.4, the 'target_url' parameter to the "
            "'Publisher.publish' method is now a string rather than a "
            "werkzeug.urls.URL instance.  To ease the transition, some "
            "methods and attributes of werkzeugs.urls.URL are being emulated, "
            "however that will not last forever. The plugin should be updated "
            "to treat 'target_url' as a string.",
            category=DeprecationWarning,
            stacklevel=2,
        )
        return rv



"""MarkdownController implementation for mistune 2.x"""

from __future__ import annotations

from collections.abc import Callable
from collections.abc import Sequence
from importlib import import_module
from typing import ClassVar
from typing import TypedDict

import mistune.util

from lektor.markdown.controller import MarkdownController
from lektor.markdown.controller import RendererHelper
from lektor.markdown.controller import UnknownPluginError
from lektor.utils import unique_everseen


def escape(text: str) -> str:
    # This is only here to provide the implementation for the
    # deprecated lektor.markdown.escape method.
    #
    # (We don't use it below and it can be deleted once access
    # to lektor.markdown.escape is removed.)
    return mistune.util.escape(text, quote=True)


class ImprovedRenderer(mistune.HTMLRenderer):  # type: ignore[misc]
    lektor: ClassVar = RendererHelper()

    # The deprecated .record and .meta attributes are not made available here, since old
    # renderer mixins (written for Lektor<3.4 and mistune 0.x) are not going to work
    # with mistune 2.x anyway.

    def link(self, link: str, text: str | None = None, title: str | None = None) -> str:
        link = self.lektor.resolve_url(link)
        return super().link(link, text, title)

    def image(self, src: str, alt: str = "", title: str | None = None) -> str:
        src = self.lektor.resolve_url(src)
        return super().image(src, alt, title)


class ParserConfigDict(TypedDict, total=False):
    block: mistune.BlockParser
    inline: mistune.InlineParser
    plugins: Sequence[Callable[[mistune.Markdown], None]]


MistunePlugin = Callable[[mistune.Markdown], None]


class MarkdownConfig:
    # Enabling these plugins give us feature parity with mistune 0.8.4.
    # We enable them by default for the sake of backwards-compatibility.
    DEFAULT_PLUGINS = ("url", "strikethrough", "footnotes", "table")

    def __init__(self) -> None:
        self.renderer_options = {
            "escape": False,
            "allow_harmful_protocols": True,
        }
        self.renderer_base = ImprovedRenderer
        self.renderer_mixins: list[type] = []
        self.parser_options: ParserConfigDict = {
            "plugins": list(self.DEFAULT_PLUGINS),
        }

    def make_renderer(self) -> ImprovedRenderer:
        bases = tuple(self.renderer_mixins) + (self.renderer_base,)
        renderer_cls = type("renderer_cls", bases, {})
        return renderer_cls(**self.renderer_options)


class MarkdownController2(MarkdownController):
    def make_parser(self) -> mistune.Markdown:
        env = self.env
        cfg = MarkdownConfig()
        # FIXME: call different hooks here for mistune 2?
        env.plugin_controller.emit("markdown-config", config=cfg)
        renderer = cfg.make_renderer()
        env.plugin_controller.emit(
            "markdown-lexer-config", config=cfg, renderer=renderer
        )
        parser_options = cfg.parser_options
        plugins = parser_options.get("plugins")
        if plugins:
            # Resolve and filter out duplicates
            parser_options["plugins"] = tuple(
                unique_everseen(map(self.resolve_plugin, plugins))
            )
        return mistune.Markdown(renderer, **cfg.parser_options)

    PLUGINS = {**mistune.PLUGINS}

    def resolve_plugin(self, plugin: str | MistunePlugin) -> MistunePlugin:
        if callable(plugin):
            return plugin
        if not isinstance(plugin, str):
            raise TypeError(
                f"Plugins should be specified as strings or callables, not {plugin!r}"
            )

        module_name, sep, attr = plugin.partition(":")
        if sep:
            try:
                return getattr(import_module(module_name.strip()), attr.strip())
            except (ImportError, AttributeError) as exc:
                raise UnknownPluginError(f"Can not load plugin {plugin!r}") from exc
        try:
            return self.PLUGINS[plugin]
        except KeyError as exc:
            raise UnknownPluginError(f"Unknown plugin {plugin!r}") from exc



from __future__ import annotations

import logging
import threading
import traceback
import webbrowser
from contextlib import ExitStack
from typing import NamedTuple
from typing import TYPE_CHECKING

from werkzeug.serving import is_running_from_reloader
from werkzeug.serving import run_simple

from lektor.admin import WebAdmin
from lektor.builder import Builder
from lektor.db import Database
from lektor.reporter import CliReporter
from lektor.utils import process_extra_flags
from lektor.watcher import watch_project


if TYPE_CHECKING:
    from _typeshed import StrPath

    from lektor.environment import Environment


class BackgroundBuilder(threading.Thread):
    """Run a thread to watch the project tree and rebuild when changes are noticed.

    This is a contextmanager. On entry, the watcher thread is started, on exit it is
    stopped.

    """

    def __init__(
        self,
        env: Environment,
        output_path: StrPath,
        prune: bool = True,
        verbosity: int = 0,
        extra_flags: dict[str, str] | None = None,
    ):
        threading.Thread.__init__(self)
        self.env = env
        self.output_path = output_path
        self.prune = prune
        self.verbosity = verbosity
        self.extra_flags = extra_flags

        # See https://github.com/samuelcolvin/watchfiles/pull/132
        self.stop_event = threading.Event()

    def __enter__(self) -> BackgroundBuilder:
        self.start()
        return self

    def __exit__(self, *args: object) -> None:
        self.stop_event.set()

    def build(self, update_source_info_first: bool = False) -> None:
        try:
            db = Database(self.env)
            builder = Builder(
                db.new_pad(), self.output_path, extra_flags=self.extra_flags
            )
            if update_source_info_first:
                builder.update_all_source_infos()
            builder.build_all()
            if self.prune:
                builder.prune()
        except Exception:
            traceback.print_exc()

    def run(self) -> None:
        watch = watch_project(self.env, self.output_path, stop_event=self.stop_event)
        with CliReporter(self.env, verbosity=self.verbosity):
            self.build(update_source_info_first=True)
            for _changes in watch:
                self.build()


class BindAddr(NamedTuple):
    host: str
    port: int


def browse_to_address(addr: BindAddr) -> None:
    timer = threading.Timer(1.0, webbrowser.open, (f"http://{addr.host}:{addr.port}",))
    timer.daemon = True
    timer.start()


def run_server(
    bindaddr: BindAddr,
    env: Environment,
    output_path: StrPath,
    *,
    prune: bool = True,
    verbosity: int = 0,
    lektor_dev: bool = False,
    ui_lang: str = "en",
    browse: bool = False,
    extra_flags: dict[str, str] | None = None,
) -> None:
    """This runs a server but also spawns a background process.  It's
    not safe to call this more than once per python process!
    """
    bindaddr = BindAddr._make(bindaddr)

    in_main_process = is_running_from_reloader() or not lektor_dev
    extra_flags = process_extra_flags(extra_flags)
    if lektor_dev:
        env.jinja_env.add_extension("jinja2.ext.debug")
    else:
        logging.getLogger("werkzeug").setLevel(logging.WARNING)

    app = WebAdmin(
        env,
        output_path=output_path,
        verbosity=verbosity,
        debug=lektor_dev,
        ui_lang=ui_lang,
        extra_flags=extra_flags,
    )

    if browse and not is_running_from_reloader():
        browse_to_address(bindaddr)

    with ExitStack() as stack:
        if in_main_process:
            env.plugin_controller.emit(
                "server-spawn", bindaddr=bindaddr, extra_flags=extra_flags
            )
            stack.callback(env.plugin_controller.emit, "server-stop")

            background_builder = BackgroundBuilder(
                env,
                output_path=output_path,
                prune=prune,
                verbosity=verbosity,
                extra_flags=extra_flags,
            )
            stack.enter_context(background_builder)

        run_simple(
            bindaddr.host,
            bindaddr.port,
            app,
            use_debugger=True,
            threaded=True,
            use_reloader=lektor_dev,
        )



from wsgiref.util import shift_path_info

import pytest
from werkzeug.exceptions import NotFound
from werkzeug.test import Client

from lektor.admin.webui import make_app


@pytest.fixture
def app(env, ui_lang, tmp_path, admin_path, app_prefix):
    app = make_app(env, output_path=tmp_path, ui_lang=ui_lang, admin_path=admin_path)
    if not app_prefix:
        return app

    def prefixed_app(environ, start_response):
        if shift_path_info(environ) != app_prefix:
            return NotFound()(environ, start_response)
        return app(environ, start_response)

    return prefixed_app


@pytest.mark.parametrize("ui_lang", ["en", "de"])
@pytest.mark.parametrize("admin_path", ["/_admin"])
@pytest.mark.parametrize("app_prefix", [None, "prefix"])
def test_dash(app, ui_lang, mocker, admin_path, app_prefix):
    render_template = mocker.patch(
        "lektor.admin.modules.dash.render_template",
        return_value="RENDERED",
    )
    script_name = f"/{app_prefix}" if app_prefix else ""

    resp = Client(app).get(f"{script_name}{admin_path}/edit?path=%2F")

    assert resp.get_data(True) == "RENDERED"
    assert render_template.mock_calls == [
        mocker.call(
            "dash.html",
            lektor_config={
                "admin_root": script_name + admin_path,
                "site_root": script_name,
                "lang": ui_lang,
            },
        )
    ]



import datetime
import re
import warnings

import pytest
from markupsafe import escape
from markupsafe import Markup

from lektor.context import Context
from lektor.datamodel import Field
from lektor.types.base import BadValue
from lektor.types.base import Undefined
from lektor.types.formats import MarkdownDescriptor


class DummySource:
    url_path = "/"

    @staticmethod
    def url_to(url, **kwargs):
        return url


def make_field(env, type, **options):
    return Field(env, "demo", type=env.types[type], options=options)


def test_markdown(env, pad):
    field = make_field(env, "markdown")

    source = DummySource()

    with Context(pad=pad):
        rv = field.deserialize_value("Hello **World**!", pad=pad)
        assert isinstance(rv, MarkdownDescriptor)
        rv = rv.__get__(source)
        assert rv
        assert rv.source == "Hello **World**!"
        assert escape(rv) == Markup("<p>Hello <strong>World</strong>!</p>\n")
        assert rv.meta == {}

        for val in "", None:
            rv = field.deserialize_value(val, pad=pad)
            assert isinstance(rv, MarkdownDescriptor)
            rv = rv.__get__(source)
            assert not rv
            assert rv.source == ""
            assert escape(rv) == Markup("")
            assert rv.meta == {}


def test_markdown_links(env, pad):
    field = make_field(env, "markdown")
    source = DummySource()

    def md(s):
        rv = field.deserialize_value(s, pad=pad)
        assert isinstance(rv, MarkdownDescriptor)
        return str(rv.__get__(source)).strip()

    with Context(pad=pad):
        assert md("[foo](http://example.com/)") == (
            '<p><a href="http://example.com/">foo</a></p>'
        )
        assert md("[foo](javascript:foo)") == (
            '<p><a href="javascript:foo">foo</a></p>'
        )

        img = (
            "iVBORw0KGgoAAAANSUhEUgAAAAUAAAAFCAYAAACNbyblAAAAHElEQVQI12P4"
            "//8/w38GIAXDIBKE0DHxgljNBAAO9TXL0Y4OHwAAAABJRU5ErkJggg=="
        )
        assert re.match(
            rf'<p><img src="data:image/png;base64,{img}" alt="test"\s*/?></p>\Z',
            md(f"![test](data:image/png;base64,{img})"),
        )


def test_markdown_linking(pad, builder):
    blog_index = pad.get("/blog", page_num=1)

    prog, _ = builder.build(blog_index)
    with prog.artifacts[0].open("rb") as f:
        assert (
            b'Look at my <a href="2015/12/post1/hello.txt">attachment</a>'
        ) in f.read()

    blog_post = pad.get("/blog/post1")

    prog, _ = builder.build(blog_post)
    with prog.artifacts[0].open("rb") as f:
        assert b'Look at my <a href="hello.txt">attachment</a>' in f.read()


def test_markdown_images(pad, builder):
    blog_index = pad.get("/blog", page_num=1)

    prog, _ = builder.build(blog_index)
    with prog.artifacts[0].open("rb") as f:
        assert re.search(
            rb'This is an image <img src="2015/12/post1/logo.png" alt="logo"\s*/?>.',
            f.read(),
        )

    blog_post = pad.get("/blog/post1")

    prog, _ = builder.build(blog_post)
    with prog.artifacts[0].open("rb") as f:
        assert re.search(
            rb'This is an image <img src="logo.png" alt="logo"\s*/?>.',
            f.read(),
        )


def test_markdown_warns_on_invalid_options(env):
    with pytest.warns(UserWarning) as warnings:
        make_field(env, "markdown", label="Test", resolve_links="GARBAGE")
    assert "Unrecognized value" in str(warnings[0].message)


@pytest.mark.parametrize("resolve_links", ["always", "never", "when-possible", None])
def test_markdown_does_not_warn_on_valid_options(env, resolve_links):
    with warnings.catch_warnings():
        warnings.simplefilter("error")
        field = make_field(env, "markdown", label="Test", resolve_links=resolve_links)
    assert field.options["resolve_links"] == resolve_links


def test_string(env, pad):
    field = make_field(env, "string")

    with Context(pad=pad):
        rv = field.deserialize_value("", pad=pad)
        assert rv == ""

        rv = field.deserialize_value(None, pad=pad)
        assert isinstance(rv, Undefined)

        rv = field.deserialize_value("foo\nbar", pad=pad)
        assert rv == "foo"

        rv = field.deserialize_value(" 123 ", pad=pad)
        assert rv == "123"


def test_text(env, pad):
    field = make_field(env, "text")

    with Context(pad=pad):
        rv = field.deserialize_value("", pad=pad)
        assert rv == ""

        rv = field.deserialize_value(None, pad=pad)
        assert isinstance(rv, Undefined)

        rv = field.deserialize_value("foo\nbar", pad=pad)
        assert rv == "foo\nbar"

        rv = field.deserialize_value(" 123 ", pad=pad)
        assert rv == " 123 "


def test_integer(env, pad):
    field = make_field(env, "integer")

    with Context(pad=pad):
        rv = field.deserialize_value("", pad=pad)
        assert isinstance(rv, BadValue)

        rv = field.deserialize_value(None, pad=pad)
        assert isinstance(rv, Undefined)

        rv = field.deserialize_value("42", pad=pad)
        assert rv == 42

        rv = field.deserialize_value(" 23 ", pad=pad)
        assert rv == 23


def test_float(env, pad):
    field = make_field(env, "float")

    with Context(pad=pad):
        rv = field.deserialize_value("", pad=pad)
        assert isinstance(rv, BadValue)

        rv = field.deserialize_value(None, pad=pad)
        assert isinstance(rv, Undefined)

        rv = field.deserialize_value("42", pad=pad)
        assert rv == 42.0

        rv = field.deserialize_value(" 23.0 ", pad=pad)
        assert rv == 23.0

        rv = field.deserialize_value("-23.5", pad=pad)
        assert rv == -23.5


def test_boolean(env, pad):
    field = make_field(env, "boolean")

    with Context(pad=pad):
        rv = field.deserialize_value("", pad=pad)
        assert isinstance(rv, BadValue)

        rv = field.deserialize_value(None, pad=pad)
        assert isinstance(rv, Undefined)

        for s in "true", "TRUE", "True", "1", "yes":
            rv = field.deserialize_value(s, pad=pad)
            assert rv is True

        for s in "false", "FALSE", "False", "0", "no":
            rv = field.deserialize_value(s, pad=pad)
            assert rv is False


dt = datetime.datetime


@pytest.mark.parametrize(
    "value, expected",
    [
        ("2016-04-30 01:02:03", dt(2016, 4, 30, 1, 2, 3)),
        ("1970-1-1 12:34", dt(1970, 1, 1, 12, 34)),
        ("1970-01-02 12:34", dt(1970, 1, 2, 12, 34)),
        ("2020-02-03 01:02:03", dt(2020, 2, 3, 1, 2, 3)),
    ],
)
def test_datetime_no_timezone(env, pad, value, expected):
    field = make_field(env, "datetime")
    with Context(pad=pad):
        rv = field.deserialize_value(value, pad=pad)

    assert rv.replace(tzinfo=None) == expected
    assert rv.tzinfo is None


def utc(*args):
    return datetime.datetime(*args, tzinfo=datetime.timezone.utc)


@pytest.mark.parametrize(
    "value, expected",
    [
        # Known timezone name, UTC
        ("2016-04-30 01:02:03 UTC", utc(2016, 4, 30, 1, 2, 3)),
        # Known timezone name, EST
        ("2016-04-30 01:02:03 EST", utc(2016, 4, 30, 6, 2, 3)),
        # Known location name, Asia/Seoul
        ("2016-04-30 01:02:03 Asia/Seoul", utc(2016, 4, 29, 16, 2, 3)),
        # KST - http://www.timeanddate.com/time/zones/kst
        ("2016-04-30 01:02:03 +0900", utc(2016, 4, 29, 16, 2, 3)),
        # ACST - http://www.timeanddate.com/time/zones/acst
        ("2016-04-30 01:02:03 +0930", utc(2016, 4, 29, 15, 32, 3)),
        # MST - http://www.timeanddate.com/time/zones/mst
        ("2016-04-30 01:02:03 -0700", utc(2016, 4, 30, 8, 2, 3)),
        # MART - http://www.timeanddate.com/time/zones/mart
        ("2016-04-30 01:02:03 -0930", utc(2016, 4, 30, 10, 32, 3)),
        # with (ignored) timezone name (case 1)
        ("2016-04-30 01:02:03 KST +0900", utc(2016, 4, 29, 16, 2, 3)),
        # with (ignored) timezone name (case 2)
        ("2016-04-30 01:02:03 KST+0900", utc(2016, 4, 29, 16, 2, 3)),
    ],
)
def test_datetime_timezone(env, pad, value, expected):
    field = make_field(env, "datetime")
    with Context(pad=pad):
        rv = field.deserialize_value(value, pad=pad)
    assert rv.astimezone(expected.tzinfo) == expected


@pytest.mark.parametrize(
    "value",
    [
        "",
        "197",
        "1970",
        "1970-01",
        "1970-01-01",
        "1970-01-01 12",
        "1970-01-01 12.23",
        "1970-01 01:02:03",
        "1970-01-01 12:34 *0800",
        "1970-01-01 12:34 -081",
        "1970-01-01 12:34 a\\b",
        "1970-01-01 12:34 very/unknown/timezone",
        "1970-01-01 12:34 very/long/timezone" + "e" * 1024,
    ],
)
def test_datetime_invalid(env, pad, value):
    field = make_field(env, "datetime")
    with Context(pad=pad):
        rv = field.deserialize_value(value, pad=pad)
    assert isinstance(rv, BadValue)


def test_datetime_missing(env, pad):
    field = make_field(env, "datetime")
    with Context(pad=pad):
        rv = field.deserialize_value(None, pad=pad)
    assert isinstance(rv, Undefined)
    assert "Missing value" in rv._undefined_hint



from __future__ import annotations

import queue
import secrets
from collections.abc import Generator
from typing import TYPE_CHECKING

from flask import Blueprint

from lektor.admin.utils import eventstream
from lektor.reporter import reporter


if TYPE_CHECKING:
    from lektor.builder import Artifact

PING_DELAY = 1.0

bp = Blueprint("livereload", __name__)
# Use a random ID to detect reloading, which will be changed after reloaded
version_id = secrets.token_urlsafe(16)


@bp.route("/events")
@eventstream
def events() -> Generator[dict[str, str]]:
    updated_artifacts: queue.Queue[Artifact] = queue.Queue()
    with reporter.on_build_change(updated_artifacts.put):
        while True:
            yield {"type": "ping", "versionId": version_id}

            try:
                artifact = updated_artifacts.get(timeout=PING_DELAY)
            except queue.Empty:
                continue
            yield {"type": "reload", "path": artifact.artifact_name}



from pathlib import Path

import pytest

from lektor.imagetools import compute_dimensions
from lektor.imagetools import get_quality


HERE = Path(__file__).parent
EXAMPLE = HERE / "../../example"


def test_compute_dimensions():
    assert compute_dimensions(10, 10, 200, 100) == (10, 5)


@pytest.mark.parametrize(
    "image, expected",
    [
        (HERE / "exif-test-1.jpg", 85),
        (EXAMPLE / "content/logo.png", 75),
    ],
)
def test_get_quality(image, expected):
    with pytest.deprecated_call():
        assert get_quality(image) == expected



import decimal
import json
import os
import subprocess
from collections import namedtuple
from datetime import timedelta

from lektor.imagetools import Thumbnail
from lektor.imagetools import ThumbnailMode
from lektor.reporter import reporter
from lektor.utils import get_dependent_url
from lektor.utils import locate_executable
from lektor.utils import portable_popen


THUMBNAIL_FORMATS = frozenset(["jpg", "jpeg", "png"])


def _imround(x):
    """Round float pixel values like imagemagick does it."""
    return decimal.Decimal(x).to_integral(decimal.ROUND_HALF_UP)


Rescaling = namedtuple("Rescaling", ["rescale", "crop"])


class Dimensions(namedtuple("Dimensions", ["width", "height"])):
    __slots__ = ()

    def __new__(cls, width, height):
        width = int(width)
        height = int(height)

        if width < 1 or height < 1:
            raise ValueError("Invalid dimensions")

        return super().__new__(cls, width, height)

    @property
    def aspect_ratio(self):
        return float(self.width) / float(self.height)

    def _infer_dimensions(self, width, height):
        """Calculate dimensions based on aspect ratio if height, width or both
        are missing.
        """
        if width is None and height is None:
            return self

        if width is None:
            width = _imround(height * self.aspect_ratio)
        elif height is None:
            height = _imround(width / self.aspect_ratio)

        return Dimensions(width, height)

    def contains(self, other):
        """Return True if the given Dimensions can be completely enclosed by
        this Dimensions.
        """
        return self.width >= other.width and self.height >= other.height

    def fit_within(self, max_width=None, max_height=None, upscale=None):
        """Calculate resizing required to make these dimensions fit within the
        given dimensions.

        Note that resizing only occurs if upscale is enabled.

            >>> source = Dimensions(640, 480)
            >>> source.fit_within(max_width=320).rescale == Dimensions(320, 240)
            True

        :param max_width: Maximum width for the new rescaled dimensions.
        :param max_height: Maximum height for the new rescaled dimensions.
        :param upscale: Allow making the dimensions larger (default False).
        :return: Rescaling operations
        :rtype: Rescaling
        """
        if upscale is None:
            upscale = False

        max_dim = self._infer_dimensions(max_width, max_height)

        # Check if we should rescale at all
        if max_dim == self or (not upscale and max_dim.contains(self)):
            return Rescaling(self, self)

        ar = self.aspect_ratio
        rescale_dim = Dimensions(
            width=_imround(min(max_dim.width, max_dim.height * ar)),
            height=_imround(min(max_dim.height, max_dim.width / ar)),
        )

        return Rescaling(rescale=rescale_dim, crop=rescale_dim)

    def cover(self, min_width=None, min_height=None, upscale=None):
        """Calculate resizing required to make these dimensions cover the given
        dimensions.

        Note that resizing only occurs if upscale is enabled.

            >>> source = Dimensions(640, 480)
            >>> target = source.cover(240, 240)
            >>> target.rescale == Dimensions(320, 240)
            True
            >>> target.crop == Dimensions(240, 240)
            True

        :param min_width: Minimum width for the new rescaled dimensions.
        :param min_height: Minimum height for the new rescaled dimensions.
        :param upscale: Allow making the dimensions larger (default True).
        :return: Rescaling operations
        :rtype: Rescaling
        """
        if upscale is None:
            upscale = True

        min_dim = self._infer_dimensions(min_width, min_height)

        # Check if we should rescale at all
        if min_dim == self or (not upscale and min_dim.contains(self)):
            return Rescaling(self, self)

        ar = self.aspect_ratio
        rescale_dim = Dimensions(
            width=_imround(max(min_dim.width, min_dim.height * ar)),
            height=_imround(max(min_dim.height, min_dim.width / ar)),
        )

        return Rescaling(rescale=rescale_dim, crop=min_dim)

    def stretch(self, width=None, height=None, upscale=None):
        """Calculate resizing required to the given dimensions without
        considering aspect ratio.

        Note that resizing only occurs if upscale is enabled.

            >>> source = Dimensions(640, 480)
            >>> source.cover(240, 240).rescale == Dimensions(240, 240)
            True

        :param min_width: Minimum width for the new rescaled dimensions.
        :param min_height: Minimum height for the new rescaled dimensions.
        :param upscale: Allow making the dimensions larger (default True).
        :return: Rescaling operations
        :rtype: Rescaling
        """
        if upscale is None:
            upscale = True

        dim = self._infer_dimensions(width, height)

        # Check if we should rescale at all
        if dim == self or (not upscale and dim.contains(self)):
            return Rescaling(self, self)

        return Rescaling(rescale=dim, crop=dim)

    def resize(self, width=None, height=None, mode=ThumbnailMode.DEFAULT, upscale=None):
        if mode == ThumbnailMode.FIT:
            return self.fit_within(width, height, upscale)
        if mode == ThumbnailMode.CROP:
            return self.cover(width, height, upscale)
        if mode == ThumbnailMode.STRETCH:
            return self.stretch(width, height, upscale)

        raise ValueError(f'Unexpected mode "{mode!r}"')


def get_timecode(td):
    """Convert a timedelta to an ffmpeg compatible string timecode.

    A timecode has the format HH:MM:SS, with decimals if needed.
    """
    seconds = td.total_seconds()

    hours = int(seconds // 3600)
    seconds %= 3600

    minutes = int(seconds // 60)
    seconds %= 60

    str_seconds, str_decimals = str(float(seconds)).split(".")

    timecode = f"{hours:02d}:{minutes:02d}:{str_seconds.zfill(2)}"
    if str_decimals != "0":
        timecode += f".{str_decimals}"

    return timecode


def get_ffmpeg_quality(quality_percent):
    """Convert a value between 0-100 to an ffmpeg quality value (2-31).

    Note that this is only applicable to the mjpeg encoder (which is used for
    jpeg images). mjpeg values works in reverse, i.e. lower is better.
    """
    if not 0 <= quality_percent <= 100:
        raise ValueError("Video quality must be between 0 and 100")

    low, high = 2, 31
    span = high - low
    factor = float(quality_percent) / 100.0
    return int(low + round(span * (1 - factor)))


def get_suffix(seek, width, height, mode, quality):
    """Make suffix for a thumbnail that is unique to the given parameters."""
    timecode = get_timecode(seek).replace(":", "-").replace(".", "-")
    bits = [f"t{timecode}"]

    if width is not None or height is not None:
        dimension = "x".join(str(x) for x in [width, height] if x is not None)
        bits.append(dimension)

    if mode != ThumbnailMode.DEFAULT:
        bits.append(mode.value)

    if quality is not None:
        bits.append(f"q{quality}")

    return "_".join(bits)


def get_video_info(filename):
    """Read video information using ffprobe if available.

    Returns a dict with: width, height and duration.
    """
    ffprobe = locate_executable("ffprobe")
    if ffprobe is None:
        raise RuntimeError("Failed to locate ffprobe")

    proc = portable_popen(
        [
            ffprobe,
            "-v",
            "quiet",
            "-print_format",
            "json",
            "-show_format",
            "-show_streams",
            filename,
        ],
        stdout=subprocess.PIPE,
    )
    stdout, _ = proc.communicate()

    if proc.returncode != 0:
        raise RuntimeError(f"ffprobe exited with code {proc.returncode}")

    ffprobe_data = json.loads(stdout.decode("utf8"))
    info = {
        "width": None,
        "height": None,
        "duration": None,
    }

    # Try to extract total video duration
    try:
        info["duration"] = timedelta(seconds=float(ffprobe_data["format"]["duration"]))
    except (KeyError, TypeError, ValueError):
        pass

    # Try to extract width and height from the first found video stream
    for stream in ffprobe_data["streams"]:
        if stream["codec_type"] != "video":
            continue

        info["width"] = int(stream["width"])
        info["height"] = int(stream["height"])

        # We currently don't bother with multiple video streams
        break

    return info


def make_video_thumbnail(
    ctx,
    source_video,
    source_url_path,
    seek,
    *,
    width=None,
    height=None,
    mode=ThumbnailMode.DEFAULT,
    upscale=None,
    quality=None,
    format=None,
):
    if mode != ThumbnailMode.FIT and (width is None or height is None):
        msg = '"%s" mode requires both `width` and `height` to be defined.'
        raise ValueError(msg % mode.value)

    if upscale is None:
        upscale = {
            ThumbnailMode.FIT: False,
            ThumbnailMode.CROP: True,
            ThumbnailMode.STRETCH: True,
        }[mode]

    if format is None:
        format = "jpg"
    if format not in THUMBNAIL_FORMATS:
        raise ValueError(f'Invalid thumbnail format "{format}"')

    if quality is not None and format != "jpg":
        raise ValueError("The quality parameter is only supported for jpeg images")

    if seek < timedelta(0):
        raise ValueError("Seek must not be negative")

    ffmpeg = locate_executable("ffmpeg")
    if ffmpeg is None:
        raise RuntimeError("Failed to locate ffmpeg")
    info = get_video_info(source_video)

    source_dim = Dimensions(info["width"], info["height"])
    resize_dim, crop_dim = source_dim.resize(width, height, mode, upscale)

    # Construct a filename suffix unique to the given parameters
    suffix = get_suffix(seek, width, height, mode=mode, quality=quality)
    dst_url_path = get_dependent_url(source_url_path, suffix, ext=f".{format}")

    if quality is None and format == "jpg":
        quality = 95

    def build_thumbnail_artifact(artifact):
        artifact.ensure_dir()

        vfilter = (
            "thumbnail",
            f"scale={resize_dim.width}:{resize_dim.height}",
            f"crop={crop_dim.width}:{crop_dim.height}",
        )

        cmdline = [
            ffmpeg,
            "-loglevel",
            "-8",
            "-ss",
            get_timecode(seek),  # Input seeking since it's faster
            "-i",
            source_video,
            "-vf",
            ",".join(vfilter),
            "-frames:v",
            "1",
            "-qscale:v",
            str(get_ffmpeg_quality(quality)),
            artifact.dst_filename,
        ]

        reporter.report_debug_info("ffmpeg cmd line", cmdline)
        proc = portable_popen(cmdline)
        if proc.wait() != 0:
            raise RuntimeError(f"ffmpeg exited with code {proc.returncode}")

        if not os.path.exists(artifact.dst_filename):
            msg = (
                f"Unable to create video thumbnail for {source_video!r}. "
                "Maybe the seek is outside of the video duration?"
            )
            raise RuntimeError(msg)

    ctx.sub_artifact(artifact_name=dst_url_path, sources=[source_video])(
        build_thumbnail_artifact
    )

    return Thumbnail(dst_url_path, crop_dim.width, crop_dim.height)
